import pytest

from base64 import b64decode
from io import StringIO, BytesIO
from bonsai import LDIFWriter
from bonsai import LDAPEntry, LDAPModOp


def test_init_params():
    """ Test constructor parameters for LDIFWriter. """
    with pytest.raises(TypeError):
        _ = LDIFWriter("wrong")
    with pytest.raises(TypeError):
        _ = LDIFWriter(StringIO(), max_length=None)
    with pytest.raises(TypeError):
        _ = LDIFWriter(BytesIO())
    out = StringIO()
    ldif = LDIFWriter(out, 100)
    assert ldif.output_file == out
    assert ldif.max_length == 100


def test_write_entry():
    """ Test serialising an LDAP entry. """
    ent = LDAPEntry("cn=test")
    ent["cn"] = "test"
    ent["jpegPhoto"] = b"1223122130008283938282931232"
    ent["sn"] = "test😊"
    ent["sn"].append(" test2")

    with StringIO() as out:
        ldif = LDIFWriter(out, max_length=32)
        ldif.write_entry(ent)
        content = out.getvalue()
    contlines = content.split("\n")
    surnames = {
        b64decode(line.split(":: ")[1]).decode("UTF-8")
        for line in contlines
        if "sn" in line
    }
    jpeg_lines = []
    for idx, line in enumerate(contlines):
        if "jpegPhoto::" in line:
            jpeg_lines.append(line.split(" ")[1])
            jpeg_lines.append(contlines[idx + 1][1:])
    jpegPhoto = b64decode("".join(jpeg_lines))

    assert all(len(line) <= 32 for line in contlines)
    assert "dn: {0}".format(ent.dn) == contlines[0]
    assert "cn: {0}\n".format(ent["cn"][0]) in content
    assert content.count("sn:: ") == 2
    assert surnames == set(ent["sn"])
    assert content.count("jpegPhoto:: ") == 1
    assert jpegPhoto == ent["jpegPhoto"][0]


def test_write_entries():
    """ Test writing multiple entries into the same output. """
    ent0 = LDAPEntry("cn=test0")
    ent0["cn"] = "test0"
    ent1 = LDAPEntry("cn=test1")
    ent1["cn"] = "test1"
    entries = (ent0, ent1)

    with StringIO() as out:
        ldif = LDIFWriter(out)
        ldif.write_entries(entries, write_version=False)
        content = out.getvalue()

    assert "dn: {0}".format(ent0.dn) in content
    assert "dn: {0}".format(ent1.dn) in content
    assert "cn: {0}".format(ent0["cn"][0]) in content
    assert "cn: {0}".format(ent1["cn"][0]) in content
    assert "version" not in content

    with StringIO() as out:
        ldif = LDIFWriter(out)
        ldif.write_entries(entries)
        content = out.getvalue()

    assert "dn: {0}".format(ent0.dn) in content
    assert "dn: {0}".format(ent1.dn) in content
    assert "version: 1" == content.split("\n")[0]


def test_write_changes():
    """ Test writing LDIF changes of an LDAP Entry. """
    ent = LDAPEntry("cn=test")
    ent["cn"] = "test"
    ent["sn"] = ["sntest1", "sntest2"]
    ent.change_attribute("givenName", LDAPModOp.REPLACE, "test")
    ent.change_attribute("uidNumber", LDAPModOp.DELETE, 0)
    ent.change_attribute("gidNumber", LDAPModOp.DELETE)

    with StringIO() as out:
        ldif = LDIFWriter(out)
        ldif.write_changes(ent)
        content = out.getvalue()
    lines = content.split("\n")

    assert "dn: {0}".format(ent.dn) == lines.pop(0)  # First line.
    assert "changetype: modify" == lines.pop(0)  # Second line.
    assert "add: cn" in lines
    assert "cn: {0}".format(ent["cn"][0]) == lines[lines.index("add: cn") + 1]
    assert "add: sn" in lines
    assert set(ent["sn"]) == {lin.split("sn: ")[1] for lin in lines if "sn: " in lin}
    assert "replace: givenName" in lines
    assert (
        "givenName: {0}".format(ent["givenName"][0])
        == lines[lines.index("replace: givenName") + 1]
    )
    assert "delete: uidNumber" in lines
    assert "uidNumber: 0" == lines[lines.index("delete: uidNumber") + 1]
    assert "delete: gidNumber" in lines
    # Remove the key entirely.
    assert "-" == lines[lines.index("delete: gidNumber") + 1]


def test_output_file():
    """ Test output_file property. """
    out = StringIO()
    ldif = LDIFWriter(out)
    assert ldif.output_file == out
    with pytest.raises(TypeError):
        ldif.output_file = None
    out2 = StringIO()
    ldif.output_file = out2
    assert ldif.output_file == out2



import sys
from functools import partial

from tornado import gen
from tornado.ioloop import IOLoop
from tornado.concurrent import Future

from ..ldapconnection import BaseLDAPConnection, LDAPSearchScope
from ..errors import LDAPError, NotAllowedOnNonleaf

# Backwards compatibility from 3.5.
if sys.version_info.minor < 5:
    StopAsyncIteration = StopIteration

class TornadoLDAPConnection(BaseLDAPConnection):
    """
    Asynchronous LDAP connection object that works with Torando.
    It has the same methods and properties as :class:`bonsai.LDAPConnection`.

    :param LDAPClient client: a client object.
    :param ioloop: a Tornado IO loop.
    """

    def __init__(self, client, ioloop=None):
        super().__init__(client, is_async=True)
        self._ioloop = ioloop or IOLoop.instance()
        self._fileno = None
        self._timeout = None

    def _io_callback(self, fut, msg_id, fd=None, events=None):
        try:
            self._ioloop.remove_handler(self._fileno)
            res = super().get_result(msg_id)
            if res is not None:
                fut.set_result(res)
                if self._timeout is not None:
                    self._ioloop.remove_timeout(self._timeout)
            else:
                self._fileno = self.fileno()
                callback = partial(self._io_callback, fut, msg_id)
                try:
                    self._ioloop.add_handler(self._fileno, callback,
                                             IOLoop.WRITE | IOLoop.READ)
                except FileExistsError as exc:
                    if exc.errno != 17:
                        raise exc
        except LDAPError as exc:
            fut.set_exception(exc)

    def _timeout_callback(self, fut):
        self._ioloop.remove_handler(self._fileno)
        fut.set_exception(gen.TimeoutError())

    def _evaluate(self, msg_id, timeout=None):
        fut = Future()
        callback = partial(self._io_callback, fut, msg_id)
        self._fileno = self.fileno()
        try:
            self._ioloop.add_handler(self._fileno, callback,
                                     IOLoop.WRITE | IOLoop.READ)
            if timeout is not None:
                self._timeout = self._ioloop.call_later(timeout,
                                                        self._timeout_callback,
                                                        fut)
        except FileExistsError as exc:
            # Avoid concurrency problems by registering with
            # the same fileno more than once.
            if exc.errno != 17:
                raise exc
        return fut

    @gen.coroutine
    def delete(self, dname, timeout=None, recursive=False):
        try:
            res = yield super().delete(dname, timeout, recursive)
            return res
        except NotAllowedOnNonleaf as exc:
            if recursive:
                results = yield self.search(dname, LDAPSearchScope.ONELEVEL,
                                            attrlist=['1.1'], timeout=timeout)
                for res in results:
                    yield self.delete(res.dn, timeout, True)
                res = yield self.delete(dname, timeout, False)
                return res
            else:
                raise exc

    @gen.coroutine
    def _search_iter_anext(self, search_iter):
        try:
            return next(search_iter)
        except StopIteration:
            msgid = search_iter.acquire_next_page()
            if msgid is None:
                raise StopAsyncIteration
            search_iter = yield self._evaluate(msgid)
            return next(search_iter)



import os.path
import sys

import pytest

from bonsai import LDAPClient
from bonsai import LDAPEntry
from bonsai import LDAPModOp
from bonsai import LDAPDN
import bonsai.errors
from bonsai.errors import InvalidDN


@pytest.fixture
def test_entry():
    """ Create a test LDAP entry. """
    gconn = None
    entry = None

    def _create_entry(conn, dname):
        nonlocal gconn
        nonlocal entry
        entry = LDAPEntry(dname)
        entry["objectclass"] = [
            "top",
            "inetOrgPerson",
            "person",
            "organizationalPerson",
        ]
        entry["sn"] = "test"
        gconn = conn
        try:
            conn.add(entry)
        except bonsai.AlreadyExists:
            conn.delete(entry.dn)
            conn.add(entry)
        return entry

    yield _create_entry

    if gconn.closed:
        gconn = gconn.open()
    gconn.delete(entry.dn)
    gconn.close()


def test_set_get():
    """ Test LDAPEntry's SetItem, GetItem and get methods. """
    entry = LDAPEntry("cn=test")
    entry["sn"] = "Test"
    assert entry == {"dn": LDAPDN("cn=test"), "sn": ["Test"]}
    entry["givenname"] = "Test"
    assert entry.get("None") is None
    assert entry.get("GivenName") == entry["givenNAME"]
    del entry["sn"]
    with pytest.raises(KeyError):
        _ = entry["sn"]


def test_append_extend():
    """ Test append and extend methods of LDAPEntry's attribute. """
    entry = LDAPEntry("cn=test")
    entry["givenName"] = "test"
    entry["givenname"].append("test2")
    assert entry["givenname"] == ["test", "test2"]
    assert entry["givenname"][0] == "test"
    with pytest.raises(ValueError):
        entry["GivenName"].extend(["teSt", "test3"])


def test_pop():
    """ Test LDAPEntry's pop method. """
    entry = LDAPEntry("cn=test")
    entry["test"] = "test"
    with pytest.raises(TypeError):
        _ = entry.pop()
    with pytest.raises(TypeError):
        _ = entry.pop("t", 2, 3)
    with pytest.raises(KeyError):
        _ = entry.pop("t")
    assert entry.pop("test") == ["test"]
    assert entry.pop("test", None) is None


def test_popitem():
    """ Test LDAPEntry's popitem method. """
    entry = LDAPEntry("cn=test")
    entry["test"] = "test"
    entry["test2"] = "test"
    item = entry.popitem()
    assert len(item) == 2
    assert item[0] not in entry
    entry[item[0]] = item[1]
    assert entry[item[0]] == item[1]


def test_popitem_empty():
    """ Test LDAPEntry's popitem raises KeyError if it is empty. """
    entry = LDAPEntry("cn=test")
    with pytest.raises(KeyError):
        _ = entry.popitem()


def test_clear():
    """ Test LDAPEntry's clear method. """
    entry = LDAPEntry("cn=test")
    entry["sn"] = ["test1", "test2"]
    entry["gn"] = ["test3"]
    entry.clear()
    assert entry == {"dn": LDAPDN("cn=test")}
    assert entry.dn == "cn=test"


def test_items():
    """ Test LDAPEntry's items method. """
    entry = LDAPEntry("cn=test")
    entry["cn"] = "test"
    entry["sn"] = "Test"
    assert len(entry.items()) == 3
    assert ("dn", entry.dn) in entry.items()
    assert ("cn", entry["cn"]) in entry.items()
    assert ("sn", entry["sn"]) in entry.items()
    assert len(list(entry.items(exclude_dn=True))) == 2
    assert ("dn", entry.dn) not in entry.items(exclude_dn=True)
    assert ("cn", entry["cn"]) in entry.items(exclude_dn=True)
    assert ("sn", entry["sn"]) in entry.items(exclude_dn=True)


def test_keys():
    """ Test LDAPEntry's keys method. """
    entry = LDAPEntry("cn=test")
    entry["cn"] = "test"
    entry["sn"] = "Test"
    assert set(entry.keys()) == set(["dn", "cn", "sn"])
    assert set(entry.keys(exclude_dn=True)) == set(["cn", "sn"])


def test_values():
    """ Test LDAPEntry's values method. """
    entry = LDAPEntry("cn=test")
    entry["cn"] = "test"
    entry["sn"] = "Test"
    assert len(entry.values()) == 3
    assert entry.dn in entry.values()
    assert entry["cn"] in entry.values()
    assert entry["sn"] in entry.values()
    assert len(list(entry.values(exclude_dn=True))) == 2
    assert entry.dn not in entry.values(exclude_dn=True)
    assert entry["cn"] in entry.values(exclude_dn=True)
    assert entry["sn"] in entry.values(exclude_dn=True)


def test_update():
    """ Test updating LDAPEntry object. """
    entry = LDAPEntry("cn=test")
    entry.update({"GivenName": "test2", "mail": "test@mail"})
    entry.update([("sn", "test")])
    with pytest.raises(ValueError):
        entry.update([("sn", "test", 1)])
    entry.update(uidnumber=1, gidnumber=1)
    assert entry["mail"] == ["test@mail"]
    assert entry["givenname"] == ["test2"]
    assert entry["sn"][0] == "test"
    assert entry["uidnumber"] == [1]
    assert entry["gidnumber"] == [1]


def test_equal():
    """ Test equality check. """
    entry1 = LDAPEntry("cn=test")
    entry2 = LDAPEntry("cn=test")
    entry3 = LDAPEntry("cn=test1")
    assert entry1 == entry2
    assert not (entry1 == entry3)
    assert entry1 == {"dn": LDAPDN("cn=test")}
    assert not (entry1 == 2)


def test_special_char(client, basedn):
    """ Test adding entry with special character in its DN. """
    with client.connect() as conn:
        entry = LDAPEntry(r"cn=test\, *\+withspec,%s" % basedn)
        entry["objectclass"] = ["top", "inetOrgPerson"]
        entry["sn"] = "Test,*special"
        conn.add(entry)
        result = conn.search(basedn, 1)
        entry.delete()
        assert entry.dn in [res.dn for res in result]


def test_unicode(client, basedn):
    """ Test adding entry with special character in its DN. """
    with client.connect() as conn:
        dname = "cn=test_µčབྷñ,%s" % basedn
        entry = LDAPEntry(dname)
        entry["objectclass"] = ["top", "inetOrgPerson"]
        entry["sn"] = "unicode_µčབྷñ"
        conn.add(entry)
        result = conn.search(dname, 0)
        entry.delete()
        assert dname in [res.dn for res in result]


def test_binary(client, basedn):
    """ Test adding binary data. """
    curdir = os.path.abspath(os.path.dirname(__file__))
    conn = client.connect()
    dname = "cn=binary,%s" % basedn
    entry = LDAPEntry(dname)
    entry["objectclass"] = ["top", "inetOrgPerson"]
    entry["sn"] = "binary_test"
    with open("%s/testenv/test.jpeg" % curdir, "rb") as image:
        entry["jpegPhoto"] = image.read()
    conn.add(entry)
    result = conn.search(dname, 0)
    entry.delete()
    conn.close()
    assert "jpegPhoto" in result[0].keys()
    assert result[0]["jpegphoto"][0] == entry["jpegphoto"][0]


def test_connection(client, basedn):
    """ Test set and get connection object from LDAPEntry. """
    entry = LDAPEntry("cn=test,%s" % basedn)
    with pytest.raises(ValueError):
        _ = entry.connection
    conn = client.connect()
    entry.connection = conn
    assert entry.connection == conn
    with pytest.raises(TypeError):
        entry.connection = "string"
    with pytest.raises(TypeError):
        del entry.connection


def test_rename(client, basedn, test_entry):
    """ Test LDAPEntry's rename LDAP operation. """
    with client.connect() as conn:
        entry = test_entry(conn, "cn=test,%s" % basedn)
        entry.rename("cn=test2,%s" % basedn)
        assert str(entry.dn) == "cn=test2,%s" % basedn
        obj = conn.search("cn=test,%s" % basedn, 0)
        assert obj == []
        obj = conn.search("cn=test2,%s" % basedn, 0)[0]
        assert entry.dn == obj.dn


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Cannot rename entry with old RDN on Windows"
)
def test_rename_with_old_rdn(client, basedn, test_entry):
    """ Test LDAPEntry's rename LDAP operation. """
    with client.connect() as conn:
        entry = test_entry(conn, "cn=test,%s" % basedn)
        entry.rename("uid=test2,%s" % basedn, delete_old_rdn=False)
        assert str(entry.dn) == "uid=test2,%s" % basedn
        obj = conn.search("cn=test,%s" % basedn, 0)
        assert obj == []
        obj = conn.search("uid=test2,%s" % basedn, 0)[0]
        assert entry.dn == obj.dn
        assert "test2" in obj["uid"]
        assert "test" in obj["cn"]


def test_rename_error(client, basedn, test_entry):
    """ Test LDAPEntry's rename error handling. """
    dname = bonsai.LDAPDN("cn=test,%s" % basedn)
    entry = LDAPEntry(dname)
    with pytest.raises(ValueError):
        entry.rename("cn=test2")
    with client.connect() as conn:
        entry = test_entry(conn, dname)
        with pytest.raises(TypeError):
            entry.rename(0)
        with pytest.raises(TypeError):
            entry.rename(0, delete_old_rdn=0)
        try:
            newdn = bonsai.LDAPDN("cn=test2,ou=invalid,%s" % basedn)
            entry.rename(newdn)
        except bonsai.LDAPError:
            assert entry.dn == dname
    with pytest.raises(bonsai.errors.ClosedConnection):
        entry.rename("cn=test2")


def test_sync_operations(client, basedn):
    """
    Test LDAPEntry's add, modify and delete synchronous operations.
    """
    entry = LDAPEntry("cn=test,%s" % basedn)
    with pytest.raises(ValueError):
        entry.modify()
    with client.connect() as conn:
        entry["sn"] = "test"
        with pytest.raises(bonsai.ObjectClassViolation):
            conn.add(entry)
        entry["objectclass"] = [
            "top",
            "inetOrgPerson",
            "person",
            "organizationalPerson",
        ]
        try:
            conn.add(entry)
        except bonsai.AlreadyExists:
            conn.delete(entry.dn)
            conn.add(entry)
        except:
            pytest.fail("Adding LDAPEntry to the server is failed.")
        entry["sn"] = "Test_modify"
        try:
            entry.modify()
        except:
            pytest.fail("Modify failed.")
        obj = conn.search("cn=test,%s" % basedn, 0)[0]
        assert entry["sn"] == obj["sn"]
        try:
            entry.delete()
        except:
            pytest.fail("Delete failed.")
    with pytest.raises(bonsai.errors.ClosedConnection):
        entry.modify()


def test_dn_attr(basedn):
    """ Test LDAPEntry's DN attribute. """
    entry = LDAPEntry("cn=test,%s" % basedn)
    entry.dn = "cn=test"
    assert str(entry.dn) == "cn=test"
    with pytest.raises(TypeError):
        del entry.dn
    with pytest.raises(TypeError):
        entry["dn"] = 5


def test_wrong_params():
    """ Test passing wrong params to LDAPEntry. """
    with pytest.raises(TypeError):
        _ = LDAPEntry("", 1)
    with pytest.raises(InvalidDN):
        _ = LDAPEntry("5", 1)


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Cannot use password policy on Windows"
)
def test_password_modify(client):
    """
    Test modifying password with simple modify operation and
    password policy.
    """
    cli = LDAPClient(client.url)
    user_dn = "cn=jeff,ou=nerdherd,dc=bonsai,dc=test"
    cli.set_password_policy(True)
    cli.set_credentials("SIMPLE", user_dn, "p@ssword")
    conn, _ = cli.connect()
    entry = conn.search(user_dn, 0)[0]
    try:
        entry["userPassword"] = "newpassword"
        entry.modify()
    except Exception as exc:
        assert isinstance(exc, bonsai.errors.PasswordModNotAllowed)
    user_dn = "cn=skip,ou=nerdherd,dc=bonsai,dc=test"
    cli.set_credentials("SIMPLE", user_dn, "p@ssword")
    conn, _ = cli.connect()
    entry = conn.search(user_dn, 0)[0]
    try:
        entry["userPassword"] = "short"
        entry.modify()
    except Exception as exc:
        assert isinstance(exc, bonsai.errors.PasswordTooShort)
    try:
        entry["userPassword"] = "p@ssword"
        entry.modify()
    except Exception as exc:
        assert isinstance(exc, bonsai.errors.PasswordInHistory)


def test_extended_dn_attr():
    """ Test that extended dn attribute is read only. """
    entry = LDAPEntry("cn=test")
    with pytest.raises(ValueError):
        entry.extended_dn = "cn=test2"


def test_change_attribute(client):
    """ Test change_attribute method. """
    user_dn = "cn=sam,ou=nerdherd,dc=bonsai,dc=test"
    if sys.platform == "win32":
        multiattr = "otherTelephone"
    else:
        multiattr = "mail"
    with client.connect() as conn:
        entry = LDAPEntry(user_dn, conn)
        entry.change_attribute("mail", LDAPModOp.ADD, "sam@bonsai.test")
        assert entry["mail"].status == 1
        entry.modify()
        assert conn.search(user_dn, 0)[0]["mail"][0] == "sam@bonsai.test"
        entry.change_attribute("mail", 1, "sam@bonsai.test")
        assert entry["mail"].status == 1
        entry.modify()
        with pytest.raises(KeyError):
            _ = conn.search(user_dn, 0)[0]["mail"]
        entry.change_attribute(
            multiattr, LDAPModOp.REPLACE, "sam@bonsai.test", "x@bonsai.test"
        )
        assert entry[multiattr].status == 2
        entry.modify()
        res = conn.search(user_dn, 0)[0][multiattr]
        assert "sam@bonsai.test" in res
        assert "x@bonsai.test" in res
        entry.change_attribute(multiattr, 1, "x@bonsai.test")
        entry.change_attribute(multiattr, 0, "sam2@bonsai.test")
        entry.modify()
        res = conn.search(user_dn, 0)[0][multiattr]
        assert "sam@bonsai.test" in res
        assert "sam2@bonsai.test" in res
        entry.change_attribute(multiattr, 1)
        entry.modify()
        assert multiattr not in conn.search(user_dn, 0)[0].keys()


def test_change_attribute_error(client):
    """ Test change_attribute method's error handling. """
    user_dn = "cn=sam,ou=nerdherd,dc=bonsai,dc=test"
    with client.connect() as conn:
        entry = LDAPEntry(user_dn, conn)
        with pytest.raises(ValueError):
            entry.change_attribute("mail", 4, "t")
        entry.change_attribute("sn", 0, "Lembeck")
        with pytest.raises(bonsai.TypeOrValueExists):
            entry.modify()
        entry.clear_attribute_changes("sn")
        entry.change_attribute("description", LDAPModOp.DELETE, "sam")
        with pytest.raises(bonsai.NoSuchAttribute):
            entry.modify()


def test_clear_attribute_changes():
    """ Test clear_attribute_changes method. """
    user_dn = "cn=sam,ou=nerdherd,dc=bonsai,dc=test"
    entry = LDAPEntry(user_dn)
    entry.change_attribute("uidNumber", 0, 4)
    assert entry["uidNumber"].added == [4]
    entry.change_attribute("uidNumber", 1, 4)
    assert entry["uidNumber"].deleted == [4]
    entry.clear_attribute_changes("uidNumber")
    assert entry["uidNumber"].status == 0
    assert entry["uidNumber"].added == []
    assert entry["uidNumber"].deleted == []


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Cannot use ManageDsaIT on Windows"
)
def test_modify_referrals(client):
    """ Test modifying an LDAP referral with ManageDdsIT control. """
    refdn = bonsai.LDAPDN("o=invalid-ref,ou=nerdherd,dc=bonsai,dc=test")
    newref = "ldap://invalid.host/cn=nobody"
    cli = LDAPClient(client.url)
    cli.set_credentials(client.mechanism, **client.credentials)
    cli.managedsait = True
    with cli.connect() as conn:
        entry = LDAPEntry(refdn, conn)
        entry.change_attribute("ref", LDAPModOp.ADD, newref)
        entry.modify()
        res = conn.search(refdn, 0, attrlist=["ref"])[0]
        assert len(res["ref"]) == 3
        assert newref in res["ref"]
        entry.change_attribute("ref", LDAPModOp.DELETE, newref)
        entry.modify()



from abc import ABCMeta, abstractmethod
from enum import IntEnum
from typing import Union, Any, List, Iterator, Tuple, Optional

from ._bonsai import ldapconnection, ldapsearchiter
from .ldapdn import LDAPDN
from .ldapentry import LDAPEntry
from .errors import UnwillingToPerform, NotAllowedOnNonleaf

MYPY = False

if MYPY:
    from .ldapclient import LDAPClient


class LDAPSearchScope(IntEnum):
    """ Enumeration for LDAP search scopes. """

    BASE = 0  #: For searching only the base DN.
    ONELEVEL = 1  #: For searching one tree level under the base DN.
    ONE = ONELEVEL  #: Alias for :attr:`LDAPSearchScope.ONELEVEL`.
    SUBTREE = 2  #: For searching the entire subtree, including the base DN.
    SUB = SUBTREE  #: Alias for :attr:`LDAPSearchScope.SUBTREE`.


class BaseLDAPConnection(ldapconnection, metaclass=ABCMeta):
    def __init__(self, client: "LDAPClient", is_async: bool = False) -> None:
        self.__client = client
        super().__init__(client, is_async)

    def __enter__(self) -> "BaseLDAPConnection":
        """ Context manager entry point. """
        return self

    def __exit__(self, type, value, traceback) -> None:
        """ Context manager exit point. """
        self.close()

    def add(self, entry: LDAPEntry, timeout: Optional[float] = None) -> Any:
        return self._evaluate(super().add(entry), timeout)

    def delete(
        self,
        dname: Union[str, LDAPDN],
        timeout: Optional[float] = None,
        recursive: bool = False,
    ) -> Any:
        if type(dname) == LDAPDN:
            dname = str(dname)
        return self._evaluate(super().delete(dname, recursive), timeout)

    def open(self, timeout: Optional[float] = None) -> Any:
        return self._evaluate(super().open(), timeout)

    def modify_password(
        self,
        user: Optional[Union[str, LDAPDN]] = None,
        new_password: Optional[str] = None,
        old_password: Optional[str] = None,
        timeout: Optional[float] = None,
    ) -> Any:
        if type(user) == LDAPDN:
            user = str(user)
        return self._evaluate(
            super().modify_password(user, new_password, old_password), timeout
        )

    def __base_search(
        self,
        base: Optional[Union[str, LDAPDN]] = None,
        scope: Optional[Union[LDAPSearchScope, int]] = None,
        filter_exp: Optional[str] = None,
        attrlist: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        sizelimit: int = 0,
        attrsonly: bool = False,
        sort_order: Optional[List[str]] = None,
        page_size: int = 0,
        offset: int = 0,
        before_count: int = 0,
        after_count: int = 0,
        est_list_count: int = 0,
        attrvalue: Optional[str] = None,
    ) -> Any:

        _base = str(base) if base is not None else str(self.__client.url.basedn)
        _scope = scope if scope is not None else self.__client.url.scope_num
        _filter = filter_exp if filter_exp is not None else self.__client.url.filter_exp
        _attrlist = attrlist if attrlist is not None else self.__client.url.attributes
        _timeout = timeout if timeout is not None else 0.0
        if sort_order is not None:
            _sort_order = self.__create_sort_list(sort_order)
        else:
            _sort_order = []
        msg_id = super().search(
            _base,
            _scope,
            _filter,
            _attrlist,
            _timeout,
            sizelimit,
            attrsonly,
            _sort_order,
            page_size,
            offset,
            before_count,
            after_count,
            est_list_count,
            attrvalue,
        )
        return self._evaluate(msg_id, timeout)

    @staticmethod
    def __create_sort_list(sort_list: List[str]) -> List[Tuple[str, bool]]:
        """
        Set a list of attribute names to sort entries in a search result. For
        reverse order set '-' before to the attribute name.

        :param list sort_list: List of attribute names.
        :raises ValueError: if any element of the list is not a string or an \
        empty string, and if any of the attributes is in the list more then \
        once.
        """
        sort_attrs = []
        for attr in sort_list:
            if not isinstance(attr, str) or len(attr) == 0:
                raise ValueError(
                    "All element of sort_list must be" " a non empty string."
                )
            if attr[0] == "-":
                # Set reverse order.
                sort_attrs.append((attr[1:], True))
            else:
                sort_attrs.append((attr, False))
        if len(sort_list) > len(set(map(lambda x: x[0].lower, sort_attrs))):
            raise ValueError("Attribute names must be different" " from each other.")
        return sort_attrs

    def search(
        self,
        base: Optional[Union[str, LDAPDN]] = None,
        scope: Optional[Union[LDAPSearchScope, int]] = None,
        filter_exp: Optional[str] = None,
        attrlist: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        sizelimit: int = 0,
        attrsonly: bool = False,
        sort_order: Optional[List[str]] = None,
    ) -> Any:
        return self.__base_search(
            base, scope, filter_exp, attrlist, timeout, sizelimit, attrsonly, sort_order
        )

    def paged_search(
        self,
        base: Optional[Union[str, LDAPDN]] = None,
        scope: Optional[Union[LDAPSearchScope, int]] = None,
        filter_exp: Optional[str] = None,
        attrlist: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        sizelimit: int = 0,
        attrsonly: bool = False,
        sort_order: Optional[List[str]] = None,
        page_size: int = 1,
    ) -> Any:
        return self.__base_search(
            base,
            scope,
            filter_exp,
            attrlist,
            timeout,
            sizelimit,
            attrsonly,
            sort_order,
            page_size,
        )

    def virtual_list_search(
        self,
        base: Optional[Union[str, LDAPDN]] = None,
        scope: Optional[Union[LDAPSearchScope, int]] = None,
        filter_exp: Optional[str] = None,
        attrlist: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        sizelimit: int = 0,
        attrsonly: bool = False,
        sort_order: Optional[List[str]] = None,
        offset: int = 1,
        before_count: int = 0,
        after_count: int = 0,
        est_list_count: int = 0,
        attrvalue: Optional[str] = None,
    ) -> Any:
        if sort_order is None and (offset != 0 or attrvalue is not None):
            raise UnwillingToPerform(
                "Sort control is required with" " virtual list view."
            )
        return self.__base_search(
            base,
            scope,
            filter_exp,
            attrlist,
            timeout,
            sizelimit,
            attrsonly,
            sort_order,
            0,
            offset,
            before_count,
            after_count,
            est_list_count,
            attrvalue,
        )

    def whoami(self, timeout: Optional[float] = None) -> Any:
        return self._evaluate(super().whoami(), timeout)

    @abstractmethod
    def _evaluate(self, msg_id: int, timeout: Optional[float] = None) -> Any:
        pass


class LDAPConnection(BaseLDAPConnection):
    """
    Handles synchronous connection to an LDAP server.

    :param LDAPClient client: a client object.
    """

    def __init__(self, client: "LDAPClient") -> None:
        super().__init__(client, False)

    def _evaluate(self, msg_id: int, timeout: Optional[float] = None) -> Any:
        """
        It returns the result of the LDAP operation.

        :param int msg_id: the ID of the LDAP operation.
        :param float timeout: time limit in seconds for the operation.
        :return: the result of the operation.
        """
        return self.get_result(msg_id, timeout)

    def add(self, entry: LDAPEntry, timeout: Optional[float] = None) -> bool:
        """
        Add new entry to the directory server.

        :param LDAPEntry entry: the new entry.
        :param float timeout: time limit in seconds for the operation.
        :return: True, if the operation is finished.
        :rtype: bool
        """
        return super().add(entry, timeout)

    def delete(
        self,
        dname: Union[str, LDAPDN],
        timeout: Optional[float] = None,
        recursive: bool = False,
    ) -> bool:
        """
        Remove entry from the directory server.

        :param str|LDAPDN dname: the string or LDAPDN format of the \
        entry's DN.
        :param float timeout: time limit in seconds for the operation.
        :param bool recursive: remove every entry of the given subtree \
        recursively.
        :return: True, if the operation is finished.
        :rtype: bool
        """
        try:
            return super().delete(dname, timeout, recursive)
        except NotAllowedOnNonleaf as exc:
            if recursive:
                results = self.search(
                    dname, LDAPSearchScope.ONELEVEL, attrlist=["1.1"], timeout=timeout
                )
                for res in results:
                    self.delete(res.dn, timeout, True)
                return self.delete(dname, timeout, False)
            else:
                raise exc

    def open(self, timeout: Optional[float] = None) -> "LDAPConnection":
        """
        Open the LDAP connection.

        :param float timeout: time limit in seconds for the operation.
        :return: The :class:`LDAPConnection` object itself.
        :rtype: :class:`LDAPConnection`.
        """
        return super().open(timeout)

    def search(
        self,
        base: Optional[Union[str, LDAPDN]] = None,
        scope: Optional[Union[LDAPSearchScope, int]] = None,
        filter_exp: Optional[str] = None,
        attrlist: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        sizelimit: int = 0,
        attrsonly: bool = False,
        sort_order: Optional[List[str]] = None,
    ) -> List[LDAPEntry]:
        # Documentation in the docs/api.rst with detailed examples.
        # Load values from the LDAPURL, if it is not presented on the
        # parameter list.
        return super().search(
            base, scope, filter_exp, attrlist, timeout, sizelimit, attrsonly, sort_order
        )

    def paged_search(
        self,
        base: Optional[Union[str, LDAPDN]] = None,
        scope: Optional[Union[LDAPSearchScope, int]] = None,
        filter_exp: Optional[str] = None,
        attrlist: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        sizelimit: int = 0,
        attrsonly: bool = False,
        sort_order: Optional[List[str]] = None,
        page_size: int = 1,
    ) -> ldapsearchiter:
        return super().paged_search(
            base,
            scope,
            filter_exp,
            attrlist,
            timeout,
            sizelimit,
            attrsonly,
            sort_order,
            page_size,
        )

    def virtual_list_search(
        self,
        base: Optional[Union[str, LDAPDN]] = None,
        scope: Optional[Union[LDAPSearchScope, int]] = None,
        filter_exp: Optional[str] = None,
        attrlist: Optional[List[str]] = None,
        timeout: Optional[float] = None,
        sizelimit: int = 0,
        attrsonly: bool = False,
        sort_order: Optional[List[str]] = None,
        offset: int = 1,
        before_count: int = 0,
        after_count: int = 0,
        est_list_count: int = 0,
        attrvalue: Optional[str] = None,
    ) -> Tuple[List[LDAPEntry], dict]:
        return super().virtual_list_search(
            base,
            scope,
            filter_exp,
            attrlist,
            timeout,
            sizelimit,
            attrsonly,
            sort_order,
            offset,
            before_count,
            after_count,
            est_list_count,
            attrvalue,
        )

    def modify_password(
        self,
        user: Optional[Union[str, LDAPDN]] = None,
        new_password: Optional[str] = None,
        old_password: Optional[str] = None,
        timeout: Optional[float] = None,
    ) -> Optional[str]:
        """
        Set a new password for the given user.

        :param str|LDAPDN user: the identification of the user. If not set, \
        the owner of the current LDAP session will be associated.
        :param str new_password: the new password. If not set, the server \
        will generate one and the new password will be returned by this method.
        :param str old_password: the current password of the user.
        :param float timeout: time limit in seconds for the operation.

        :return: if the `new_password` is not set, then the generated \
        password, None otherwise.
        :rtype: str|None
        """
        return super().modify_password(user, new_password, old_password, timeout)

    def whoami(self, timeout: Optional[float] = None) -> str:
        """
        This method can be used to obtain authorization identity.

        :param float timeout: time limit in seconds for the operation.

        :return: the authorization ID.
        :rtype: str
        """
        return super().whoami(timeout)



import re
from typing import Union, Tuple, Any

from .errors import InvalidDN


class LDAPDN:
    """
    A class for handling valid LDAP distinguished name.

    :param str strdn: a string representation of LDAP distinguished name.
    """

    __slots__ = ("__strdn",)

    _attrtype = r"[A-Za-z][\w-]*|\d+(?:\.\d+)*"
    _attrvalue = r'#(?:[\dA-Fa-f]{2})+|(?:[^,=\+<>#;\\"]|\\[,=\+<>#;\\" ]' r'|\\[\dA-Fa-f]{2})*|"(?:[^\\"]|\\[,=\+<>#;\\"]|\\[\dA-Fa-f]{2})*"'
    _namecomp = r"({typ})=({val})(?:\+({typ})=({val}))*".format(
        typ=_attrtype, val=_attrvalue
    )
    _dnregex = re.compile(
        r"({comp})(?:,({comp}))*\Z".format(comp=_namecomp), re.IGNORECASE
    )

    def __init__(self, strdn: str) -> None:
        if strdn != "" and not self._dnregex.match(strdn):
            raise InvalidDN(strdn)
        self.__strdn = self.__sanitize(strdn, True)

    def __str_rdn_to_tuple(self, str_rdn: str) -> Tuple[Tuple[str, str], ...]:
        """
        Generate attribute type and value tuple from relative
        distinguished name.
        """
        rdn = []
        # Split relative distinguished name to attribute types and values.
        type_value_list = re.split(r"(?<!\\)[+]", str_rdn)
        for attr in type_value_list:
            # Get attribute type and value.
            atype, avalue = re.split(r"(?<!\\)=", attr)
            rdn.append((atype, self.__sanitize(avalue, True)))
        return tuple(rdn)

    @staticmethod
    def __sanitize(strdn: str, reverse: bool = False) -> str:
        """ Sanitizing special characters."""
        char_list = [
            (r"\\", "\\5C"),
            (r"\,", "\\2C"),
            (r"\+", "\\2B"),
            (r"\"", "\\22"),
            (r"\<", "\\3C"),
            (r"\>", "\\3E"),
            (r"\;", "\\3B"),
            (r"\=", "\\3D"),
            (r"\ ", "\\20"),
        ]
        if strdn:
            for pair in char_list:
                if reverse:
                    strdn = strdn.replace(pair[1], pair[0])
                else:
                    strdn = strdn.replace(pair[0], pair[1])
        return strdn

    def __getitem__(self, idx: Union[int, slice]) -> str:
        """
        Return the string format of the relative distinguished names
        in the LDAPDN.

        :param int idx: the indices of the RDNs.
        :return: the string format of the RDNs.
        :rtype: str
        """
        rdns = re.split(r"(?<!\\),", self.__strdn)
        if isinstance(idx, int):
            if idx >= len(rdns):
                raise IndexError("Index is out of range.")
            # Create a slice to avoid join string characters.
            idx = slice(idx, idx + 1)
        elif not isinstance(idx, slice):
            raise TypeError("Indices must be integers or slices.")
        return ",".join(rdns[idx])

    def __setitem__(self, idx: Union[int, slice], value: str) -> None:
        """
        Set the string format of the relative distinguished names
        in the LDAPDN.

        :param int idx: the indices of the RDNs.
        :param str value: the new RDNs.
        """
        if not isinstance(value, str):
            raise ValueError("New value must be string.")
        if isinstance(idx, int):
            idx = slice(idx, idx + 1)
        elif not isinstance(idx, slice):
            raise TypeError("Indices must be integers or slices.")
        if not self._dnregex.match(value):
            raise InvalidDN(value)
        rdns = re.split(r"(?<!\\),", self.__strdn)
        rdns[idx] = re.split(r"(?<!\\),", value)
        self.__strdn = ",".join(rdns)

    def __eq__(self, other: Any) -> bool:
        """
        Check equality of two LDAPDN by their string format or
        their sanitized string format.
        """
        return (
            str(self).lower() == str(other).lower()
            or self.__sanitize(str(self)).lower() == self.__sanitize(str(other)).lower()
        )

    def __str__(self) -> str:
        """ Return the full string format of the distinguished name. """
        return self.__strdn

    def __len__(self) -> int:
        """ Return the number of RDNs of the distinguished name. """
        return len(re.split(r"(?<!\\),", self.__strdn))

    def __repr__(self) -> str:
        """ The representation of LDAPDN class. """
        return "<LDAPDN %s>" % str(self)

    @property
    def rdns(self) -> Tuple[Tuple[Tuple[str, str], ...], ...]:
        """ The tuple of relative distinguished name."""
        return tuple(
            self.__str_rdn_to_tuple(rdn)
            for rdn in re.split(r"(?<!\\),", self.__sanitize(self.__strdn))
        )

    @rdns.setter
    def rdns(self, value: Any = None) -> None:
        """ The tuple of relative distinguished names."""
        raise ValueError("RDNs attribute cannot be set.")



import configparser
import os
import time
from contextlib import contextmanager

import pytest

from bonsai import LDAPClient


def get_config():
    """ Load config parameters. """
    curdir = os.path.abspath(os.path.dirname(__file__))
    cfg = configparser.ConfigParser()
    cfg.read(os.path.join(curdir, "test.ini"))
    return cfg


@contextmanager
def network_delay(delay):
    import xmlrpc.client as rpc

    ipaddr = get_config()["SERVER"]["hostip"]
    proxy = rpc.ServerProxy("http://%s:%d/" % (ipaddr, 8000))
    proxy.set_delay(delay)
    time.sleep(2.0)
    try:
        yield proxy
    finally:
        proxy.remove_delay()


@pytest.fixture(scope="module")
def client():
    """ Get an LDAPClient with simple authentication. """
    cfg = get_config()
    url = "ldap://%s:%s/%s??%s" % (
        cfg["SERVER"]["hostip"],
        cfg["SERVER"]["port"],
        cfg["SERVER"]["basedn"],
        cfg["SERVER"]["search_scope"],
    )
    cli = LDAPClient(url)
    cli.set_credentials(
        "SIMPLE", user=cfg["SIMPLEAUTH"]["user"], password=cfg["SIMPLEAUTH"]["password"]
    )
    return cli


@pytest.fixture(scope="module")
def basedn():
    """ Get base DN. """
    cfg = get_config()
    return cfg["SERVER"]["basedn"]



import asyncio

from ..pool import ConnectionPool, ClosedPool, EmptyPool

from .aioconnection import AIOLDAPConnection

MYPY = False

if MYPY:
    from ..ldapclient import LDAPClient


class AIOPoolContextManager:
    def __init__(self, pool, *args, **kwargs):
        self.pool = pool
        self.__conn = None

    async def __aenter__(self):
        if self.pool.closed:
            await self.pool.open()
        self.__conn = await self.pool.get()
        return self.__conn

    async def __aexit__(self, type, value, traceback):
        await self.pool.put(self.__conn)

class AIOConnectionPool(ConnectionPool):
    """
    A connection pool that can be used with asnycio tasks. It's inherited from 
    :class:`bonsai.pool.ConnectionPool`.

    :param LDAPClient client: the :class:`bonsai.LDAPClient` that's used to create
                connections.
    :param int minconn: the minimum number of connections that's created
                after the pool is opened.
    :param int maxconn: the maximum number of connections in the pool.
    :param \*\*kwargs: additional keyword arguments that are passed to
                the :meth:`bonsai.LDAPClient.connect` method.
    :raises ValueError: when the minconn is negative or the maxconn is less
        than the minconn.
    """
    def __init__(
        self,
        client: "LDAPClient",
        minconn: int = 1,
        maxconn: int = 10,
        loop=None,
        **kwargs
    ):
        super().__init__(client, minconn, maxconn, **kwargs)
        self._loop = loop
        self._lock = asyncio.Condition(loop=self._loop)

    async def open(self):
        for _ in range(self._minconn):
            conn = await self._client.connect(
                is_async=True, loop=self._loop, **self._kwargs
            )
            self._idles.add(conn)
        self._closed = False

    async def get(self) -> AIOLDAPConnection:
        async with self._lock:
            if self._closed:
                raise ClosedPool("The pool is closed.")
            await self._lock.wait_for(lambda: not self.empty or self._closed)
            try:
                conn = self._idles.pop()
            except KeyError:
                if len(self._used) < self._maxconn:
                    conn = await self._client.connect(
                        is_async=True, loop=self._loop, **self._kwargs
                    )
                else:
                    raise EmptyPool("Pool is empty.") from None
            self._used.add(conn)
            self._lock.notify()
            return conn

    async def put(self, conn: AIOLDAPConnection) -> None:
        async with self._lock:
            super().put(conn)
            self._lock.notify()

    async def close(self):
        async with self._lock:
            super().close()
            self._lock.notify_all()

    def spawn(self, *args, **kwargs):
        return AIOPoolContextManager(self, *args, **kwargs)



import os
import sys
import subprocess
import tempfile
import time

import pytest
from conftest import get_config, network_delay

import bonsai
from bonsai import LDAPDN
from bonsai import LDAPClient
from bonsai import LDAPConnection
from bonsai import LDAPSearchScope
import bonsai.errors
from bonsai.errors import ClosedConnection
from bonsai.ldapconnection import BaseLDAPConnection


class SimpleAsyncConn(BaseLDAPConnection):
    def __init__(self, client):
        super().__init__(client, True)

    def _evaluate(self, msg_id, timeout=None):
        return msg_id


@pytest.fixture(scope="module")
def cfg():
    """ Get config. """
    return get_config()


@pytest.fixture(scope="module")
def ipaddr():
    """ Get IP address. """
    cfg = get_config()
    return cfg["SERVER"]["hostip"]


@pytest.fixture(scope="module")
def ktpath():
    """ Get keytab file path. """
    cfg = get_config()
    proj_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))
    ktpath = os.path.abspath(os.path.join(proj_dir, cfg["GSSAPIAUTH"]["keytab"]))
    return ktpath


@pytest.fixture
def binding():
    """ Create a binding with the server. """

    def _create_binding(auth, mech, authzid=None, realm=None):
        cfg = get_config()
        host = "ldap://%s" % cfg["SERVER"]["hostname"]
        client = LDAPClient(host)
        client.set_credentials(
            mech, cfg[auth]["user"], cfg[auth]["password"], realm, authzid
        )
        return client.connect()

    return _create_binding


@pytest.fixture
def kinit():
    def _create_kinit(authzid=None):
        try:
            cfg = get_config()
            user = cfg["GSSAPIAUTH"]["user"]
            password = cfg["GSSAPIAUTH"]["password"]
            proc = subprocess.Popen(
                ["kinit", "--version"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            output = " ".join(proc.communicate())
            if "Heimdal" in output:
                # Heimdal Kerberos implementation.
                with tempfile.NamedTemporaryFile() as psw_tmp:
                    psw_tmp.write(password.encode())
                    psw_tmp.flush()
                    cmd = ["kinit", "--password-file=%s" % psw_tmp.name, user]
                    subprocess.check_call(cmd)
            else:
                # MIT Kerberos implementation.
                cmd = 'echo "%s" | kinit %s' % (password, user)
                subprocess.check_output(cmd, shell=True)
            host = "ldap://%s" % cfg["SERVER"]["hostname"]
            client = LDAPClient(host)
            client.set_credentials(
                "GSSAPI",
                cfg["GSSAPIAUTH"]["user"],
                cfg["GSSAPIAUTH"]["password"],
                None,
                authzid,
            )
            return client.connect()
        except subprocess.CalledProcessError:
            pytest.fail("Receiving TGT is failed.")

    yield _create_kinit
    subprocess.call("kdestroy")


@pytest.fixture
def external_binding():
    """ Create a binding using external mechanism. """

    def _create_external(authzid=None):
        tls_impl = bonsai.get_tls_impl_name()
        if tls_impl == "GnuTLS" or tls_impl == "OpenSSL":
            cfg = get_config()
            curdir = os.path.abspath(os.path.dirname(__file__))
            cert_path = os.path.join(curdir, "testenv", "certs")
            host = "ldap://%s" % cfg["SERVER"]["hostname"]
            cli = LDAPClient(host, tls=True)
            cli.set_ca_cert(cert_path + "/cacert.pem")
            cli.set_client_cert(cert_path + "/client.pem")
            cli.set_client_key(cert_path + "/client.key")
            cli.set_credentials("EXTERNAL", authz_id=authzid)
            return cli.connect()
        else:
            pytest.skip("")

    return _create_external


def _generate_client(cfg):
    url = "ldap://{host}:{port}/ou=nerdherd,{basedn}?{attr}?{scope}".format(
        host=cfg["SERVER"]["hostip"],
        port=cfg["SERVER"]["port"],
        basedn=cfg["SERVER"]["basedn"],
        attr=cfg["SERVER"]["search_attr"],
        scope=cfg["SERVER"]["search_scope"],
    )
    client = LDAPClient(url)
    client.set_credentials(
        "SIMPLE", user=cfg["SIMPLEAUTH"]["user"], password=cfg["SIMPLEAUTH"]["password"]
    )
    client.auto_page_acquire = False
    return client


@pytest.fixture
def conn():
    """ Create a connection. """
    cfg = get_config()
    client = _generate_client(cfg)
    return client.connect()


@pytest.fixture
def async_conn():
    cfg = get_config()
    client = _generate_client(cfg)
    return SimpleAsyncConn(client)


def test_bind_digest(binding):
    """ Test DIGEST-MD5 connection. """
    with binding("DIGESTAUTH", "DIGEST-MD5") as conn:
        assert "anonymous" != conn.whoami()


@pytest.mark.skipif(
    sys.platform.startswith("win") or get_config()["DIGESTAUTH"]["authzid"] == "None",
    reason="Authzid is not set",
)
def test_bind_digest_with_authzid(binding, cfg):
    """ Test DIGEST-MD5 connection with authorization ID. """
    authzid = cfg["DIGESTAUTH"]["authzid"]
    with binding("DIGESTAUTH", "DIGEST-MD5", authzid) as conn:
        assert cfg["DIGESTAUTH"]["dn"] == conn.whoami()


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="NTLM is not enabled on Windows."
)
def test_bind_ntlm(binding):
    """ Test NTLM connection. """
    with binding("NTLMAUTH", "NTLM") as conn:
        assert "anonymous" != conn.whoami()


def test_bind_not_supported_auth(binding):
    """ Test not supported authentication mechanism by the server. """
    with pytest.raises(bonsai.AuthMethodNotSupported):
        _ = binding("DIGESTAUTH", "SCRAM-SHA-1")


@pytest.mark.skipif(sys.platform != "linux", reason="Kinit only available on Linux.")
def test_bind_gssapi_kinit(kinit):
    """ Test GSSAPI connection. """
    with kinit() as conn:
        assert "anonymous" != conn.whoami()


@pytest.mark.skipif(sys.platform != "linux", reason="Kinit only available on Linux.")
def test_bind_gssapi_with_authzid_kinit(kinit, cfg):
    """ Test GSSAPI connection with authorization ID. """
    authzid = cfg["GSSAPIAUTH"]["authzid"]
    with kinit(authzid) as conn:
        assert cfg["GSSAPIAUTH"]["dn"] == conn.whoami()


@pytest.mark.skipif(
    not bonsai.has_krb5_support() or sys.platform == "darwin",
    reason="Module doesn't have KRB5 support.",
)
def test_bind_gssapi(binding, cfg):
    """ Test GSSAPI connection with automatic TGT requesting. """
    with binding(
        "GSSAPIAUTH", "GSSAPI", None, cfg["GSSAPIAUTH"]["realm"].upper()
    ) as conn:
        assert "anonymous" != conn.whoami()


@pytest.mark.skipif(
    not bonsai.has_krb5_support() or sys.platform == "darwin",
    reason="Module doesn't have KRB5 support.",
)
def test_bind_gssapi_error(cfg):
    """ Test automatic TGT requesting with wrong realm name. """
    client = _generate_client(cfg)
    client.set_credentials(
        "GSSAPI",
        cfg["GSSAPIAUTH"]["user"],
        cfg["GSSAPIAUTH"]["password"],
        cfg["GSSAPIAUTH"]["realm"],
        None,
    )
    with pytest.raises(bonsai.AuthenticationError):
        _ = client.connect()


@pytest.mark.skipif(
    not bonsai.has_krb5_support() or sys.platform != "linux",
    reason="Keytab-based auth only available on linux.",
)
def test_bind_gssapi_keytab_error(cfg, ktpath):
    client = LDAPClient("ldap://%s" % cfg["SERVER"]["hostname"])
    client.set_credentials(
        "GSSAPI",
        user=cfg["GSSAPIAUTH"]["user"],
        realm=cfg["GSSAPIAUTH"]["realm"].upper(),
        keytab="invalid",
    )
    with pytest.raises(bonsai.AuthenticationError):
        _ = client.connect()
    client.set_credentials(
        "GSSAPI",
        user="invalid",
        realm=cfg["GSSAPIAUTH"]["realm"].upper(),
        keytab=ktpath,
    )
    with pytest.raises(bonsai.AuthenticationError):
        _ = client.connect()


@pytest.mark.skipif(
    not bonsai.has_krb5_support() or sys.platform != "linux",
    reason="Keytab-based auth only available on linux.",
)
def test_bind_gssapi_keytab(cfg, ktpath):
    client = LDAPClient("ldap://%s" % cfg["SERVER"]["hostname"])
    client.set_credentials(
        "GSSAPI",
        user=cfg["GSSAPIAUTH"]["user"],
        realm=cfg["GSSAPIAUTH"]["realm"].upper(),
        keytab=ktpath,
    )
    conn = client.connect()
    assert conn.whoami() == "dn:cn=chuck,ou=nerdherd,dc=bonsai,dc=test"
    client.set_credentials(
        "GSSAPI", user="admin", realm=cfg["GSSAPIAUTH"]["realm"].upper(), keytab=ktpath
    )
    conn = client.connect()
    assert conn.whoami() == "dn:cn=admin,dc=bonsai,dc=test"


@pytest.mark.skipif("TRAVIS" in os.environ, reason="No GSS-SPNEGO mech on Trusty")
def test_bind_spnego(binding, cfg):
    """ Test GSS-SPNEGO connection with automatic TGT requesting. """
    with binding(
        "GSSAPIAUTH", "GSS-SPNEGO", None, cfg["GSSAPIAUTH"]["realm"].upper()
    ) as conn:
        assert "anonymous" != conn.whoami()


def test_bind_external(external_binding):
    """ Test EXTERNAL connection. """
    with external_binding() as conn:
        assert "anonymous" != conn.whoami()


def test_bind_external_with_authzid(external_binding, cfg):
    """ Test EXTERNAL connection with authorization ID. """
    authzid = cfg["EXTERNALAUTH"]["authzid"]
    with external_binding(authzid) as conn:
        assert cfg["EXTERNALAUTH"]["dn"] == conn.whoami()


def test_search(conn, basedn):
    """ Test searching. """
    res = conn.search("ou=nerdherd,%s" % basedn, LDAPSearchScope.SUB)
    assert res is not None
    assert len(res) == len(conn.search())


def test_search_ldapdn(conn, basedn):
    """ Test searching with LDAPDN object. """
    ldap_dn = LDAPDN(basedn)
    obj = conn.search(ldap_dn, 1)
    assert obj is not None


def test_search_attr(conn, basedn):
    """ Test searching with given list of attributes. """
    obj = conn.search(basedn, 2, "(objectclass=person)", ["cn"])[0]
    assert obj is not None
    assert "cn" in obj.keys()


def test_search_attrsonly(conn, basedn):
    """ Test search receiving only attributes. """
    obj = conn.search(basedn, 2, "(objectclass=person)", ["cn"], attrsonly=True)[0]
    assert obj is not None
    assert obj["cn"] == []


def test_add_and_delete(conn, basedn):
    """ Test adding and removing an LDAP entry. """
    entry = bonsai.LDAPEntry("cn=example,%s" % basedn)
    entry.update(
        {"objectclass": ["top", "inetorgperson"], "cn": "example", "sn": "example"}
    )
    try:
        conn.add(entry)
        res = conn.search(entry.dn, 0)
        assert res[0] == entry
        conn.delete("cn=example,%s" % basedn)
        res = conn.search(entry.dn, 0)
        assert res == []
        with pytest.raises(ValueError):
            conn.add(bonsai.LDAPEntry(""))
    except bonsai.LDAPError:
        pytest.fail("Add and delete new entry is failed.")


def test_recursive_delete(conn, basedn):
    """ Test removing a subtree recursively. """
    org1 = bonsai.LDAPEntry("ou=testusers,%s" % basedn)
    org1.update({"objectclass": ["organizationalUnit", "top"], "ou": "testusers"})
    org2 = bonsai.LDAPEntry("ou=tops,ou=testusers,%s" % basedn)
    org2.update({"objectclass": ["organizationalUnit", "top"], "ou": "tops"})
    entry = bonsai.LDAPEntry("cn=tester,ou=tops,ou=testusers,%s" % basedn)
    entry.update(
        {"objectclass": ["top", "inetorgperson"], "cn": "tester", "sn": "example"}
    )
    try:
        conn.add(org1)
        conn.add(org2)
        conn.add(entry)
        conn.delete(org1.dn, recursive=True)
        res = conn.search(org1.dn, 2)
        assert res == []
    except bonsai.LDAPError:
        pytest.fail("Recursive delete is failed.")


def test_whoami(conn, cfg):
    """ Test whoami. """
    obj = conn.whoami()
    expected_res = [
        "dn:%s" % cfg["SIMPLEAUTH"]["user"],
        cfg["SIMPLEAUTH"]["adusername"],
    ]
    assert obj in expected_res


def test_connection_error():
    """ Test connection error. """
    client = LDAPClient("ldap://invalid")
    with pytest.raises(bonsai.ConnectionError):
        _ = client.connect()


def test_simple_auth_error(cfg):
    """ Test simple authentication error. """
    client = LDAPClient("ldap://%s" % cfg["SERVER"]["hostname"])
    client.set_credentials("SIMPLE", "cn=wrong", "wronger")
    with pytest.raises(bonsai.AuthenticationError):
        _ = client.connect()


def test_digest_auth_error(cfg):
    """ Test DIGEST-MD5 authentication error. """
    client = LDAPClient("ldap://%s" % cfg["SERVER"]["hostname"])
    if cfg["DIGESTAUTH"]["realm"] == "None":
        realm = None
    else:
        realm = cfg["DIGESTAUTH"]["realm"].upper()
    client.set_credentials(
        "DIGEST-MD5", cfg["DIGESTAUTH"]["user"], "wrongpassword", realm, None
    )
    with pytest.raises(bonsai.AuthenticationError):
        _ = client.connect()


def test_sort_order(conn, basedn):
    """ Test setting sort order. """
    obj = conn.search(basedn, 2, attrlist=["uidNumber"], sort_order=["-uidNumber"])
    sort = [o["uidNumber"][0] for o in obj if "uidNumber" in o]
    assert all(sort[i] >= sort[i + 1] for i in range(len(sort) - 1))


def test_fileno(conn, cfg):
    """ Test fileno method. """
    assert isinstance(conn.fileno(), int)
    try:
        import socket

        sock = socket.fromfd(conn.fileno(), socket.AF_INET, socket.SOCK_RAW)
        assert sock.getpeername() == (
            cfg["SERVER"]["hostip"],
            int(cfg["SERVER"]["port"]),
        )
        sock.close()
    except OSError:
        pytest.fail("Not a valid socket descriptor.")


def test_close(conn):
    """ Test close method. """
    conn.close()
    assert conn.closed
    with pytest.raises(bonsai.ClosedConnection):
        _ = conn.whoami()


def test_abandon(async_conn, basedn):
    """ Test abandon method. """
    msgid = async_conn.open()
    while async_conn.get_result(msgid) is None:
        pass
    msgid = async_conn.search(basedn, 2)
    async_conn.abandon(msgid)
    with pytest.raises(bonsai.InvalidMessageID):
        _ = async_conn.get_result(msgid)


def test_async_close_remove_pendig_ops(async_conn, basedn):
    """ Test remove pending operations after close. """
    msgid = async_conn.open()
    while async_conn.get_result(msgid) is None:
        pass
    async_conn.search(basedn, 2)
    async_conn.search(basedn, 0)
    async_conn.close()
    assert async_conn.closed


def test_vlv_offset(conn, basedn):
    """ Test VLV control with offset. """
    search_dn = "ou=nerdherd,%s" % basedn
    res, ctrl = conn.virtual_list_search(
        search_dn,
        1,
        attrlist=["uidNumber"],
        offset=2,
        sort_order=["-uidNumber"],
        before_count=1,
        after_count=1,
        est_list_count=6,
    )
    assert len(res) == 3
    assert ctrl["target_position"] == 2
    assert ctrl["list_count"] == 6
    assert res[1]["uidNumber"][0] == 4


def test_vlv_attrvalue(conn, basedn):
    """ Test VLV control with attribute value. """
    search_dn = "ou=nerdherd,%s" % basedn
    res, ctrl = conn.virtual_list_search(
        search_dn,
        1,
        attrlist=["uidNumber"],
        attrvalue=2,
        sort_order=["uidNumber"],
        before_count=1,
        after_count=2,
        est_list_count=6,
    )
    assert len(res) == 4
    assert ctrl["target_position"] == 3
    assert res[0]["uidNumber"][0] == 1


def test_vlv_without_sort_order(conn, basedn):
    """ Test VLV control without sort control. """
    search_dn = "ou=nerdherd,%s" % basedn
    with pytest.raises(bonsai.UnwillingToPerform):
        _ = conn.virtual_list_search(
            search_dn,
            1,
            attrlist=["uidNumber"],
            offset=1,
            before_count=1,
            after_count=2,
            est_list_count=6,
        )


def test_paged_search(conn, basedn):
    """ Test paged results control. """
    search_dn = "ou=nerdherd,%s" % basedn
    res = conn.paged_search(search_dn, 1, page_size=2)
    for ent in res:
        assert isinstance(ent, bonsai.LDAPEntry)
    page = 1  # First page is already acquired.
    while True:
        if len(res) > 2:
            pytest.fail("The size of the page is greater than expected.")
        msgid = res.acquire_next_page()
        if msgid is None:
            break
        res = conn.get_result(msgid)
        page += 1
    assert page == 3


def test_paged_search_with_auto_acq(cfg, basedn):
    """ Test paged results control with automatic page acquiring. """
    client = LDAPClient("ldap://%s" % cfg["SERVER"]["hostname"])
    conn = client.connect()
    search_dn = "ou=nerdherd,%s" % basedn
    res = conn.paged_search(search_dn, 1, page_size=3)
    if len(res) != 3:
        pytest.fail("The size of the page is not what is expected.")
    entry = 0
    for ent in res:
        assert isinstance(ent, bonsai.LDAPEntry)
        entry += 1
    assert entry == 6
    assert res.acquire_next_page() is None


@pytest.mark.timeout(15)
def test_search_timeout(conn, basedn):
    """ Test search method's timeout. """
    search_dn = "ou=nerdherd,%s" % basedn
    with pytest.raises(TypeError):
        _ = conn.search(search_dn, 1, timeout=True)
    with pytest.raises(ValueError):
        _ = conn.search(search_dn, 1, timeout=-15)
    with network_delay(6.1):
        with pytest.raises(bonsai.TimeoutError):
            _ = conn.search(search_dn, 1, timeout=3.0)


@pytest.mark.timeout(10)
def test_whoami_timeout(conn):
    """ Test whoami's timeout. """
    with pytest.raises(TypeError):
        _ = conn.whoami(timeout="A")
    with pytest.raises(ValueError):
        _ = conn.whoami(timeout=-10)
    with pytest.raises(bonsai.TimeoutError):
        _ = conn.whoami(timeout=0)
    with network_delay(6.1):
        with pytest.raises(bonsai.TimeoutError):
            _ = conn.whoami(timeout=3.2)


def test_wrong_conn_param():
    """ Test passing wrong parameters for LDAPConnection. """
    with pytest.raises(TypeError):
        _ = LDAPConnection("wrong")
    with pytest.raises(TypeError):
        _ = LDAPConnection(1)


def test_wrong_search_param(ipaddr):
    """ Test passing wrong parameters for search method. """
    with pytest.raises(ClosedConnection):
        cli = LDAPClient("ldap://%s" % ipaddr)
        LDAPConnection(cli).search()
    with pytest.raises(ValueError):
        cli = LDAPClient("ldap://%s" % ipaddr)
        LDAPConnection(cli).open().search()
    with pytest.raises(TypeError):
        cli = LDAPClient("ldap://%s" % ipaddr)
        LDAPConnection(cli).open().search("", 0, 3)


def test_wrong_add_param(conn, ipaddr):
    """ Test passing wrong parameter for add method. """
    with pytest.raises(ClosedConnection):
        cli = LDAPClient("ldap://%s" % ipaddr)
        LDAPConnection(cli).add(bonsai.LDAPEntry("cn=dummy"))
    with pytest.raises(TypeError):
        conn.add("wrong")


def test_wrong_delete_param(conn, ipaddr):
    """ Test passing wrong parameter for delete method. """
    with pytest.raises(ClosedConnection):
        cli = LDAPClient("ldap://%s" % ipaddr)
        LDAPConnection(cli).delete("cn=dummy")
    with pytest.raises(TypeError):
        conn.delete(0)


@pytest.mark.skipif(
    sys.platform == "win32", reason="Cannot use password policy on Windows."
)
def test_password_lockout(conn, ipaddr):
    """ Test password locking with password policy. """
    user_dn = "cn=jeff,ou=nerdherd,dc=bonsai,dc=test"
    cli = LDAPClient("ldap://%s" % ipaddr)
    cli.set_password_policy(True)
    try:
        cli.set_credentials("SIMPLE", user_dn, "wrong_pass")
        test_conn, ctrl = cli.connect()
    except bonsai.errors.AuthenticationError:
        with pytest.raises(bonsai.errors.AccountLocked):
            cli.set_credentials("SIMPLE", user_dn, "p@ssword")
            test_conn, ctrl = cli.connect()
    finally:
        entry = conn.search(user_dn, 0, attrlist=["pwdAccountLockedTime"])[0]
        if "pwdAccountLockedTime" in entry.keys():
            del entry["pwdAccountLockedTime"]
            entry.modify()


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Cannot use password policy on Windows."
)
def test_password_expire(conn, ipaddr):
    """ Test password expiring with password policy. """
    user_dn = "cn=skip,ou=nerdherd,dc=bonsai,dc=test"
    cli = LDAPClient("ldap://%s" % ipaddr)
    cli.set_password_policy(True)
    cli.set_credentials("SIMPLE", user_dn, "p@ssword")
    test_conn, ctrl = cli.connect()
    entry = test_conn.search(user_dn, 0)[0]
    entry["userPassword"] = "newvalidpassword"
    entry.modify()
    test_conn.close()
    cli.set_credentials("SIMPLE", user_dn, "newvalidpassword")
    time.sleep(2.0)
    test_conn, ctrl = cli.connect()
    if not (ctrl["expire"] <= 10 and ctrl["expire"] > 0):
        pytest.fail("Expire time is in the wrong range (Expire: %d)." % ctrl["expire"])
    test_conn.close()
    time.sleep(10)
    test_conn, ctrl = cli.connect()
    assert ctrl["grace"] == 1
    test_conn.close()
    with pytest.raises(bonsai.errors.PasswordExpired):
        test_conn, ctrl = cli.connect()
    entry = conn.search(user_dn, 0, attrlist=["userPassword"])[0]
    entry["userPassword"] = "p@ssword"
    entry.modify()
    entry = conn.search(user_dn, 0, attrlist=["pwdChangeTime", "pwdGraceUseTime"])[0]
    if ("pwdChangeTime", "pwdGraceUseTime") in entry.keys():
        del entry["pwdChangeTime"]
        del entry["pwdGraceUseTime"]
        entry.modify()


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="Cannot use password modify extended operation on Windows.",
)
def test_password_modify_extop(conn, ipaddr):
    """ Test Password Modify extended operation. """
    user_dn = LDAPDN("cn=skip,ou=nerdherd,dc=bonsai,dc=test")
    cli = LDAPClient("ldap://%s" % ipaddr)
    cli.set_credentials("SIMPLE", str(user_dn), "p@ssword")
    test_conn = cli.connect()
    with pytest.raises(TypeError):
        test_conn.modify_password(new_password=0)
    test_conn.modify_password(user_dn, "newpassword", "p@ssword")
    test_conn.close()
    with pytest.raises(ClosedConnection):
        test_conn.modify_password()
    try:
        cli.set_credentials("SIMPLE", str(user_dn), "newpassword")
        cli.set_password_policy(True)
        test_conn, ctrl = cli.connect()
        newpass = test_conn.modify_password()
        test_conn.close()
        assert isinstance(newpass, str)
        cli.set_credentials("SIMPLE", str(user_dn), newpass)
        test_conn, ctrl = cli.connect()
        test_conn.close()
    except bonsai.AuthenticationError:
        pytest.fail("Failed to authenticate with the new password.")
    finally:
        entry = conn.search(user_dn, 0, attrlist=["userPassword"])[0]
        entry["userPassword"] = "p@ssword"
        entry.modify()
        entry = conn.search(user_dn, 0, attrlist=["pwdChangeTime", "pwdGraceUseTime"])[
            0
        ]
        if ("pwdChangeTime", "pwdGraceUseTime") in entry.keys():
            del entry["pwdChangeTime"]
            del entry["pwdGraceUseTime"]
            entry.modify()


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Cannot use ManageDsaIT on Windows"
)
def test_search_with_managedsait_ctrl(ipaddr):
    """ Test searching with manageDsaIT control. """
    refdn = LDAPDN("o=admin-ref,ou=nerdherd,dc=bonsai,dc=test")
    cli = LDAPClient("ldap://%s" % ipaddr)
    with cli.connect() as conn:
        res = conn.search(refdn, LDAPSearchScope.BASE, attrlist=["ref"])[0]
        assert str(res.dn) == "cn=admin,dc=bonsai,dc=test"
    cli.set_managedsait(True)
    with cli.connect() as conn:
        res = conn.search(refdn, LDAPSearchScope.BASE, attrlist=["ref"])[0]
        assert refdn == res.dn
        assert "ldap://bonsai.test/cn=admin,dc=bonsai,dc=test" == res["ref"][0]


@pytest.mark.skipif(
    sys.platform.startswith("win"), reason="Cannot use ManageDsaIT on Windows"
)
def test_add_and_delete_referrals(cfg, ipaddr):
    """ Test add and delete an LDAP referral with ManageDdsIT control. """
    refdn = bonsai.LDAPDN("o=test-ref,ou=nerdherd,dc=bonsai,dc=test")
    ref = "ldap://test.host/cn=nobody"
    cli = LDAPClient("ldap://%s" % ipaddr)
    cli.set_credentials(
        "SIMPLE", user=cfg["SIMPLEAUTH"]["user"], password=cfg["SIMPLEAUTH"]["password"]
    )
    cli.managedsait = True
    with cli.connect() as conn:
        entry = bonsai.LDAPEntry(refdn, conn)
        entry["objectClass"] = ["referral", "extensibleObject"]
        entry["o"] = "test-ref"
        entry["ref"] = ref
        conn.add(entry)
        res = conn.search(refdn, 0, attrlist=["ref"])[0]
        assert entry.dn == res.dn
        assert entry["ref"] == res["ref"]
    cli.managedsait = False
    with cli.connect() as conn:
        with pytest.raises(bonsai.LDAPError):
            conn.delete(entry.dn)
    cli.managedsait = True
    with cli.connect() as conn:
        conn.delete(entry.dn)
        res = conn.search(refdn, 0, attrlist=["ref"])
        assert res == []



import sys

import pytest
from conftest import get_config, network_delay

import bonsai
from bonsai import LDAPClient
from bonsai.ldapconnection import LDAPConnection


@pytest.fixture(scope="module")
def url():
    """ Get an LDAPClient with simple authentication. """
    cfg = get_config()
    url = "ldap://%s:%s" % (cfg["SERVER"]["hostip"], cfg["SERVER"]["port"])
    return bonsai.LDAPURL(url)


def test_ldapurl(url):
    """ Test setting LDAPURL. """
    cli = LDAPClient(url)
    assert cli.url == url
    with pytest.raises(TypeError):
        _ = LDAPClient(None)
    cli.url = "ldap://localhost"
    assert cli.url == "ldap://localhost"


def test_connect(client):
    """ Test connect method. """
    assert client.connect() is not None


def test_rootdse(client):
    """ Test receiving root DSE. """
    root_dse = client.get_rootDSE()
    assert root_dse["supportedLDAPVersion"][0] == 3


def test_raw_attributes(client):
    """ Test setting raw attributes to keep in bytearray format. """
    with pytest.raises(TypeError):
        client.set_raw_attributes([5])
    with pytest.raises(ValueError):
        client.raw_attributes = ["ou", "cn", "ou"]
    client.set_raw_attributes(["ou"])
    conn = client.connect()
    result = conn.search("ou=nerdherd,dc=bonsai,dc=test", 0)[0]
    assert isinstance(result["ou"][0], bytes)
    assert not isinstance(result["objectClass"][0], bytes)


def test_set_credentials(url):
    """ Test set_credentials method, mechanism and credentials properties. """
    client = LDAPClient(url)
    with pytest.raises(TypeError):
        client.set_credentials(2323, user=None)
    with pytest.raises(TypeError):
        client.set_credentials("Simple", "Name", 2, None, None)
    client.set_credentials("SIMPLE", "cn=admin", "password")
    assert client.mechanism == "SIMPLE"
    assert client.credentials == {
        "user": "cn=admin",
        "password": "password",
        "realm": None,
        "authz_id": None,
        "keytab": None,
    }
    client.set_credentials("EXTERNAL", authz_id="authzid")
    assert client.credentials["authz_id"] == "authzid"


def test_vendor_info():
    """ Test vendor information. """
    info = bonsai.get_vendor_info()
    assert len(info) == 2
    assert isinstance(info[0], str)
    assert isinstance(info[1], int)


def test_tls_impl_name():
    """ Test TLS implementation name. """
    tls_impl = bonsai.get_tls_impl_name()
    assert tls_impl in ("GnuTLS", "MozNSS", "OpenSSL", "SChannel")


def test_debug():
    """ Test setting debug mode. """
    import subprocess

    code = "import bonsai; bonsai.set_debug(True); bonsai.LDAPClient('ldap://a.com').connect()"
    try:
        output = subprocess.check_output(
            [sys.executable, "-c", code], universal_newlines=True
        )
    except subprocess.CalledProcessError as exc:
        output = exc.output
    assert "DBG: ldapconnection_new " in output


@pytest.mark.timeout(18)
def test_connection_timeout(client):
    """ Test connection timeout. """
    with pytest.raises(TypeError):
        _ = client.connect(timeout="Wrong")
    with pytest.raises(ValueError):
        _ = client.connect(timeout=-1.5)
    with pytest.raises(bonsai.TimeoutError):
        _ = client.connect(timeout=0)
    with network_delay(9.0):
        with pytest.raises(bonsai.TimeoutError):
            client.connect(timeout=5.0)


def test_ppolicy(url):
    """ Test password policy setting. """
    client = LDAPClient(url)
    with pytest.raises(TypeError):
        client.set_password_policy("F")
    client.password_policy = True
    client.set_credentials(
        "SIMPLE", "cn=chuck,ou=nerdherd,dc=bonsai,dc=test", "p@ssword"
    )
    ret_val = client.connect()
    assert isinstance(ret_val, tuple)
    assert isinstance(ret_val[0], LDAPConnection)
    if ret_val[1] is None:
        pass  # Password policy is not supported.
    elif isinstance(ret_val[1], dict):
        assert "oid" in ret_val[1].keys()
        assert "expire" in ret_val[1].keys()
        assert "grace" in ret_val[1].keys()
        assert "1.3.6.1.4.1.42.2.27.8.5.1" == ret_val[1]["oid"]
    else:
        pytest.fail("Invalid second object in the tuple.")
    ret_val[0].close()


def test_extended_dn(url):
    """ Test extended dn control. """
    client = LDAPClient(url)
    with pytest.raises(TypeError):
        client.set_extended_dn("A")
    with pytest.raises(ValueError):
        client.set_extended_dn(2)
    client.extended_dn_format = 0
    assert client.extended_dn_format == 0
    conn = client.connect()
    root_dse = client.get_rootDSE()
    result = conn.search("ou=nerdherd,dc=bonsai,dc=test", 0)[0]
    if "1.2.840.113556.1.4.529" in root_dse["supportedControl"]:
        assert result.extended_dn is not None
        assert result.extended_dn.split(";")[-1] == str(result.dn)
    else:
        assert result.extended_dn is None


def test_readonly_attributes(client):
    """ Test read-only attributes of LDAPClient. """
    with pytest.raises(ValueError):
        client.mechanism = "SIMPLE"
    with pytest.raises(ValueError):
        client.credentials = {"user": "test", "password": "test"}
    with pytest.raises(ValueError):
        client.tls = False


def test_auto_acquire_prop(client):
    """ Test auto_page_acquire property. """
    with pytest.raises(TypeError):
        client.set_auto_page_acquire("A")
    assert client.auto_page_acquire
    client.auto_page_acquire = False
    assert not client.auto_page_acquire


def test_server_chase_referrals(client):
    """ Test server_chase_referrals property. """
    with pytest.raises(TypeError):
        client.set_server_chase_referrals(2)
    assert client.server_chase_referrals
    client.server_chase_referrals = False
    assert not client.server_chase_referrals


def test_managedsait(client):
    """ Test managedsait property. """
    with pytest.raises(TypeError):
        client.set_managedsait("B")
    assert not client.managedsait
    client.managedsait = True
    assert client.managedsait


@pytest.mark.skipif(
    get_config()["SERVER"]["has_tls"] == "False", reason="TLS is not set"
)
def test_tls(url):
    """ Test TLS connection. """
    client = LDAPClient(url, True)
    client.set_cert_policy("ALLOW")
    client.set_ca_cert(None)
    client.set_ca_cert_dir(None)
    try:
        conn = client.connect()
        conn.close()
    except Exception as exc:
        pytest.fail("TLS connection is failed with: %s" % str(exc))


@pytest.mark.skipif(
    get_config()["SERVER"]["has_tls"] == "False", reason="TLS is not set"
)
@pytest.mark.timeout(18)
def test_tls_timeout(url):
    """ Test TLS connection timeout. """
    client = LDAPClient(url, True)
    client.set_cert_policy("ALLOW")
    client.set_ca_cert(None)
    client.set_ca_cert_dir(None)
    with network_delay(9.0):
        with pytest.raises(bonsai.TimeoutError):
            client.connect(timeout=5.0)


@pytest.mark.skipif(sys.platform.startswith("win"), reason="No IPC support on Windows")
def test_ldapi():
    """ Test making connection via IPC. """
    client = LDAPClient("ldapi://%2Ftmp%2Fbonsai%2Fldapi")
    assert client.connect() is not None



import asyncio
import sys
import time
from functools import wraps

import pytest
from conftest import get_config, network_delay

from bonsai import LDAPClient
from bonsai import LDAPEntry
from bonsai.asyncio import AIOConnectionPool, AIOLDAPConnection
from bonsai.pool import ClosedPool
import bonsai.errors


def asyncio_test(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        future = func(*args, **kwargs)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(future)

    return wrapper


@asyncio_test
async def test_connection(client):
    """ Test opening a connection. """
    conn = await client.connect(True)
    assert conn is not None
    assert conn.closed == False


@asyncio_test
async def test_search(client):
    """ Test search. """
    async with client.connect(True) as conn:
        res = await conn.search()
        assert res is not None


@asyncio_test
async def test_add_and_delete(client, basedn):
    """ Test adding and deleting an LDAP entry. """
    async with client.connect(True) as conn:
        entry = LDAPEntry("cn=async_test,%s" % basedn)
        entry["objectclass"] = [
            "top",
            "inetOrgPerson",
            "person",
            "organizationalPerson",
        ]
        entry["sn"] = "async_test"
        try:
            await conn.add(entry)
        except bonsai.errors.AlreadyExists:
            await conn.delete(entry.dn)
            await conn.add(entry)
        except:
            pytest.fail("Unexpected error.")
        res = await conn.search()
        assert entry in res
        await entry.delete()
        res = await conn.search()
        assert entry not in res


@asyncio_test
async def test_recursive_delete(client, basedn):
    """ Test removing a subtree recursively. """
    org1 = bonsai.LDAPEntry("ou=testusers,%s" % basedn)
    org1.update({"objectclass": ["organizationalUnit", "top"], "ou": "testusers"})
    org2 = bonsai.LDAPEntry("ou=tops,ou=testusers,%s" % basedn)
    org2.update({"objectclass": ["organizationalUnit", "top"], "ou": "tops"})
    entry = bonsai.LDAPEntry("cn=tester,ou=tops,ou=testusers,%s" % basedn)
    entry.update(
        {"objectclass": ["top", "inetorgperson"], "cn": "tester", "sn": "example"}
    )
    try:
        with (await client.connect(True)) as conn:
            await conn.add(org1)
            await conn.add(org2)
            await conn.add(entry)
            with pytest.raises(bonsai.errors.NotAllowedOnNonleaf):
                await conn.delete(org1.dn)
            await conn.delete(org1.dn, recursive=True)
            res = await conn.search(org1.dn, 2)
            assert res == []
    except bonsai.LDAPError as err:
        pytest.fail("Recursive delete is failed: %s" % err)


@asyncio_test
async def test_modify_and_rename(client, basedn):
    """ Test modifying and renaming LDAP entry. """
    async with client.connect(True) as conn:
        entry = LDAPEntry("cn=async_test,%s" % basedn)
        entry["objectclass"] = [
            "top",
            "inetOrgPerson",
            "person",
            "organizationalPerson",
        ]
        entry["sn"] = "async_test"
        oldname = "cn=async_test,%s" % basedn
        newname = "cn=async_test2,%s" % basedn
        res = await conn.search(newname, 0)
        if res:
            await res[0].delete()
        try:
            await conn.add(entry)
        except bonsai.errors.AlreadyExists:
            await conn.delete(entry.dn)
            await conn.add(entry)
        except:
            pytest.fail("Unexpected error.")
        entry["sn"] = "async_test2"
        await entry.modify()
        await entry.rename(newname)
        res = await conn.search(entry.dn, 0, attrlist=["sn"])
        assert entry["sn"] == res[0]["sn"]
        res = await conn.search(oldname, 0)
        assert res == []
        await conn.delete(entry.dn)


@asyncio_test
async def test_obj_err(client, basedn):
    """ Test object class violation error. """
    entry = LDAPEntry("cn=async_test,%s" % basedn)
    entry["cn"] = ["async_test"]
    with pytest.raises(bonsai.errors.ObjectClassViolation):
        async with client.connect(True) as conn:
            await conn.add(entry)


@asyncio_test
async def test_whoami(client):
    """ Test whoami. """
    async with client.connect(True) as conn:
        cfg = get_config()
        obj = await conn.whoami()
        expected_res = [
            "dn:%s" % cfg["SIMPLEAUTH"]["user"],
            cfg["SIMPLEAUTH"]["adusername"],
        ]
        assert obj in expected_res


@pytest.mark.timeout(18)
@asyncio_test
async def test_connection_timeout(client):
    """ Test connection timeout. """
    with network_delay(6.0):
        with pytest.raises(asyncio.TimeoutError):
            await client.connect(True, timeout=8.0)


@pytest.mark.timeout(18)
@asyncio_test
async def test_search_timeout(client):
    """ Test search timeout. """
    async with client.connect(True) as conn:
        with network_delay(5.1):
            with pytest.raises(asyncio.TimeoutError):
                await conn.search(timeout=4.0)


@pytest.mark.skipif(
    sys.version_info.minor < 5, reason="No __aiter__ and __anext__ methods under 3.5."
)
@asyncio_test
async def test_paged_search(client, basedn):
    """ Test paged search. """
    search_dn = "ou=nerdherd,%s" % basedn
    async with client.connect(True) as conn:
        # To keep compatibility with 3.4 it does not uses async for,
        # but its while loop equivalent.
        res_iter = await conn.paged_search(search_dn, 1, page_size=3)
        res_iter = type(res_iter).__aiter__(res_iter)
        cnt = 0
        while True:
            try:
                res = await type(res_iter).__anext__(res_iter)
                assert isinstance(res, LDAPEntry)
                cnt += 1
            except StopAsyncIteration:
                break
        assert cnt == 6


@asyncio_test
async def test_async_with(client):
    """ Test async with context manager (with backward compatibility). """
    mgr = client.connect(True)
    aexit = type(mgr).__aexit__
    aenter = type(mgr).__aenter__(mgr)

    conn = await aenter
    try:
        assert conn.closed == False
        _ = await conn.whoami()
    except:
        if not (await aexit(mgr, *sys.exc_info())):
            raise
    else:
        await aexit(mgr, None, None, None)
    assert conn.closed


async def keep(pool, delay):
    conn = await pool.get()
    await asyncio.sleep(delay)
    await pool.put(conn)


@asyncio_test
async def test_pool_get_put(client):
    """ Test getting and putting back connection from pool. """
    delay = 2
    pool = AIOConnectionPool(client, minconn=1, maxconn=1)
    with pytest.raises(ClosedPool):
        _ = await pool.get()
    await pool.open()
    assert pool.closed == False
    assert pool.idle_connection == 1
    task1 = asyncio.ensure_future(keep(pool, delay))
    task2 = asyncio.ensure_future(keep(pool, delay))
    start = time.time()
    await task1
    await task2
    assert time.time() - start >= delay * 2
    # With enough connection in the pool for both tasks.
    pool.max_connection = 2
    task1 = asyncio.ensure_future(keep(pool, delay))
    task2 = asyncio.ensure_future(keep(pool, delay))
    start = time.time()
    await task1
    await task2
    assert time.time() - start >= delay


@asyncio_test
async def test_pool_close(client):
    """ Test closing the pool. """
    pool = AIOConnectionPool(client, minconn=1, maxconn=1)
    await pool.open()
    assert pool.closed == False
    assert pool.idle_connection == 1
    await pool.close()
    assert pool.closed == True
    assert pool.idle_connection == 0


@asyncio_test
async def test_pool_pass_param(client):
    """ Test passing parameter to connect. """
    pool = AIOConnectionPool(client, minconn=1, maxconn=1, timeout=0)
    with pytest.raises(asyncio.TimeoutError):
        await pool.open()
        _ = await pool.get()


@asyncio_test
async def test_pool_spawn(client):
    """ Test context manager. """
    pool = AIOConnectionPool(client, minconn=1, maxconn=1)
    assert pool.idle_connection == 0
    async with pool.spawn() as conn:
        assert pool.shared_connection == 1
        _ = await conn.whoami()
    assert pool.idle_connection == 1
    assert pool.shared_connection == 0


import pytest

from bonsai.utils import escape_filter_exp
from bonsai.utils import escape_attribute_value
from bonsai.utils import set_connect_async


def test_escape_attribute_value():
    """ Test escaping special characters in attribute values. """
    assert (
        escape_attribute_value(" dummy=test,something+somethingelse")
        == r"\ dummy\=test\,something\+somethingelse"
    )
    assert escape_attribute_value("#dummy=test ") == r"\#dummy\=test\ "
    assert escape_attribute_value(r"term\0") == r"term\\0"


def test_escape_filter_exp():
    """ Test escaping filter expressions. """
    assert escape_filter_exp("(parenthesis)") == "\\28parenthesis\\29"
    assert escape_filter_exp("cn=*") == "cn=\\2A"
    assert escape_filter_exp("\\backslash") == "\\5Cbackslash"
    assert escape_filter_exp("term\0") == "term\\0"


def test_set_async_connect(client):
    """ Dummy test for set_connect_async. """
    with pytest.raises(TypeError):
        set_connect_async("true")
    set_connect_async(False)
    conn = client.connect()
    assert conn is not None


from typing import Optional, Type


class LDAPError(Exception):
    """General LDAP error."""

    code = 0

    @classmethod
    def create(cls, code: int) -> Type["LDAPError"]:
        """ Create a new LDAPError type with `code` error code. """
        cls.code = code
        return cls

    @property
    def hexcode(self) -> int:
        """ Error code in 16 bit length hexadecimal format. """
        return (self.code + (1 << 16)) % (1 << 16)

    def __str__(self) -> str:
        return "{} (0x{:04X} [{:d}])".format(
            self.args[0] if self.args else "", self.hexcode, self.code
        )


class InvalidDN(LDAPError):
    """Raised, when dn string is not a valid distinguished name."""

    code = 0x22


class ConnectionError(LDAPError):
    """Raised, when client is not able to connect to the server."""

    code = -1


class AuthenticationError(LDAPError):
    """Raised, when authentication is failed with the server."""

    code = 0x31


class AuthMethodNotSupported(LDAPError):
    """Raised, when the chosen authentication method is not supported. """

    code = 0x07


class ObjectClassViolation(LDAPError):
    """Raised, when try to add or modify an LDAP entry and it violates the
    object class rules."""

    code = 0x41


class AlreadyExists(LDAPError):
    """Raised, when try to add an entry and it already exists in the
    dictionary. """

    code = 0x44


class InvalidMessageID(LDAPError):
    """Raised, when try to get the result with a message ID that belongs to an
    unpending or already finished operation."""

    code = -100


class ClosedConnection(LDAPError):
    """Raised, when try to perform LDAP operation with closed connection."""

    code = -101


class InsufficientAccess(LDAPError):
    """Raised, when the user has insufficient access rights."""

    code = 0x32


class TimeoutError(LDAPError):
    """Raised, when the specified timeout is exceeded. """

    code = -5


class ProtocolError(LDAPError):
    """Raised, when protocol error is happened."""

    code = 0x02


class UnwillingToPerform(LDAPError):
    """Raised, when the server is not willing to handle requests."""

    code = 0x35


class NoSuchObjectError(LDAPError):
    """
    Raised, when operation (except search) is performed on
    an entry that is not found in the directory.
    """

    code = 0x20


class AffectsMultipleDSA(LDAPError):
    """Raised, when multiple directory server agents are affected. """

    code = 0x47


class SizeLimitError(LDAPError):
    """Raised, when the size limit of the search is exceeded."""

    code = 0x04


class NotAllowedOnNonleaf(LDAPError):
    """Raised, when the operation is not allowed on a nonleaf object."""

    code = 0x42


class NoSuchAttribute(LDAPError):
    """Raised, when the given attribute of an entry does not exist."""

    code = 0x10


class TypeOrValueExists(LDAPError):
    """
    Raised, when the attribute already exists or the value
    has been already assigned.
    """

    code = 0x14


class PasswordPolicyError(LDAPError):
    """ General exception for password policy errors. """

    _dflt_args = ("Password policy error.",)

    def __init__(self, msg: Optional[str] = None) -> None:
        super().__init__(msg)
        self.args = self._dflt_args if msg is None else (msg,)


class PasswordExpired(PasswordPolicyError, AuthenticationError):
    """
    Raised, when the password policy is set, available on the server
    and the user's password is expired.
    """

    code = -200
    _dflt_args = ("User's password is expired.",)


class AccountLocked(PasswordPolicyError, AuthenticationError):
    """
    Raised, when the password policy is set, available on the server
    and the user's account is locked.
    """

    code = -201
    _dflt_args = ("User's account is locked.",)


class ChangeAfterReset(PasswordPolicyError):
    """
    Raised, when the password policy is set, available on the server
    and it signifies that the password must be changed before the user
    will be allowed to perform any operation (except bind and modify).
    """

    code = -202
    _dflt_args = ("User's password is expired.",)


class PasswordModNotAllowed(PasswordPolicyError):
    """
    Raised, when the password policy is set, available on the server
    and the user is restricted from changing her password.
    """

    code = -203
    _dflt_args = ("Password modification is not allowed.",)


class MustSupplyOldPassword(PasswordPolicyError):
    """
    Raised, when the password policy is set, available on the server
    and the existing password is not specified.
    """

    code = -204
    _dflt_args = ("Old password must be provided.",)


class InsufficientPasswordQuality(PasswordPolicyError):
    """
    Raised, when the password policy is set, available on the server
    and the user's password is not strong enough.
    """

    code = -205
    _dflt_args = ("Password does not pass quality checking.",)


class PasswordTooShort(PasswordPolicyError):
    """
    Raised, when the password policy is set, available on the server
    and the user's password is too short to be set.
    """

    code = -206
    _dflt_args = ("Password is too short.",)


class PasswordTooYoung(PasswordPolicyError):
    """
    Raised, when the password policy is set, available on the server
    and the user's password is too young to be modified.
    """

    code = -207
    _dflt_args = ("Password is too young to be modified.",)


class PasswordInHistory(PasswordPolicyError):
    """
    Raised, when the password policy is set, available on the server
    and the user's password is in the history.
    """

    code = -208
    _dflt_args = ("User's password is in the history.",)


def _get_error(code: int) -> type:
    """ Return an error by code number. """
    if code == -1 or code == 0x51 or code == -11:
        # WinLDAP returns 0x51 for Server Down.
        # OpenLDAP returns -11 for Connection error.
        return ConnectionError.create(code)
    elif code == 0x02:
        return ProtocolError
    elif code == 0x04:
        return SizeLimitError
    elif code == 0x07:
        return AuthMethodNotSupported
    elif code == 0x10:
        return NoSuchAttribute
    elif code == 0x14:
        return TypeOrValueExists
    elif code == 0x20:
        return NoSuchObjectError
    elif code == 0x22:
        return InvalidDN
    elif code == 0x31:
        return AuthenticationError
    elif code == 0x32:
        return InsufficientAccess
    elif code == 0x35:
        return UnwillingToPerform
    elif code == 0x41:
        return ObjectClassViolation
    elif code == 0x42:
        return NotAllowedOnNonleaf
    elif code == 0x44:
        return AlreadyExists
    elif code == 0x47:
        return AffectsMultipleDSA
    elif code == -5 or code == 0x55:
        return TimeoutError.create(code)
    elif code == -100:
        return InvalidMessageID
    elif code == -101:
        return ClosedConnection
    elif code == -200:
        return PasswordExpired
    elif code == -201:
        return AccountLocked
    elif code == -202:
        return ChangeAfterReset
    elif code == -203:
        return PasswordModNotAllowed
    elif code == -204:
        return MustSupplyOldPassword
    elif code == -205:
        return InsufficientPasswordQuality
    elif code == -206:
        return PasswordTooShort
    elif code == -207:
        return PasswordTooYoung
    elif code == -208:
        return PasswordInHistory
    else:
        return LDAPError.create(code)



import threading
from contextlib import contextmanager
from typing import Optional, Dict, Any

MYPY = False

if MYPY:
    from .ldapclient import LDAPClient


class PoolError(Exception):
    """ Connection pool related errors. """
    pass


class ClosedPool(PoolError):
    """ Raised, when the connection pool is closed. """
    pass


class EmptyPool(PoolError):
    """ Raised, when the connection pool is empty. """
    pass


class ConnectionPool:
    """
    A connection pool object for managing multiple open connections.

    :param LDAPClient client: the :class:`bonsai.LDAPClient` that's used to create
                connections.
    :param int minconn: the minimum number of connections that's created
                after the pool is opened.
    :param int maxconn: the maximum number of connections in the pool.
    :param \*\*kwargs: additional keyword arguments that are passed to
                the :meth:`bonsai.LDAPClient.connect` method.
    :raises ValueError: when the minconn is negative or the maxconn is less
        than the minconn.
    """
    def __init__(
        self,
        client: "LDAPClient",
        minconn: int = 1,
        maxconn: int = 10,
        **kwargs: Dict[str, Any]
    ) -> None:
        """ Init method. """
        if minconn < 0:
            raise ValueError("The minconn must be positive.")
        if minconn > maxconn:
            raise ValueError("The maxconn must be greater than minconn.")
        self._minconn = minconn
        self._maxconn = maxconn
        self._client = client
        self._kwargs = kwargs
        self._closed = True
        self._idles = set()
        self._used = set()

    def open(self) -> None:
        """
        Open the connection pool by initialising the minimal number of
        connections.
        """
        for _ in range(self._minconn):
            self._idles.add(self._client.connect(self._kwargs))
        self._closed = False

    def get(self):
        """
        Get a connection from the connection pool.

        :raises EmptyPool: when the pool is empty.
        :raises ClosedPool: when the method is called on a closed pool.
        :return: an LDAP connection object.
        """
        if self._closed:
            raise ClosedPool("The pool is closed.")
        try:
            conn = self._idles.pop()
        except KeyError:
            if len(self._used) < self._maxconn:
                conn = self._client.connect(self._kwargs)
            else:
                raise EmptyPool("Pool is empty.") from None
        self._used.add(conn)
        return conn

    def put(self, conn) -> None:
        """
        Put back a connection to the connection pool.

        :param LDAPConnection conn: the connection managed by the pool.
        :raises ClosedPool: when the method is called on a closed pool.
        :raises PoolError: when tying to put back an object that's not managed
                by this pool.
        """
        if self._closed:
            raise ClosedPool("The pool is closed.")
        try:
            self._used.remove(conn)
            self._idles.add(conn)
        except KeyError:
            raise PoolError("The %r is not managed by this pool." % conn) from None

    def close(self) -> None:
        """ Close the pool and all of its managed connections. """
        for conn in self._idles:
            conn.close()
        for conn in self._used:
            conn.close()
        self._closed = True
        self._idles = set()
        self._used = set()

    @contextmanager
    def spawn(self, *args, **kwargs):
        """
        Context manager method that acquires a connection from the pool
        and returns it on exit. It also opens the pool if it hasn't been
        opened before.

        :params \*args: the positional arguments passed to
                :meth:`bonsai.pool.ConnectionPool.get`.
        :params \*\*kwargs: the keyword arguments passed to
                :meth:`bonsai.pool.ConnectionPool.get`.
        """
        try:
            if self._closed:
                self.open()
            conn = self.get(*args, **kwargs)
            yield conn
        finally:
            self.put(conn)

    @property
    def empty(self) -> bool:
        """
        Read-only property that will be True when the connection pool has
        no free connection to use.
        """
        return len(self._idles) == 0 and len(self._used) == self._maxconn

    @property
    def closed(self) -> bool:
        """
        Read-only property that will be True when the connection pool has
        been closed.
        """
        return self._closed

    @property
    def shared_connection(self) -> int:
        """ The number of shared connections. """
        return len(self._used)

    @property
    def idle_connection(self) -> int:
        """ the number of idle connection. """
        return len(self._idles)

    @property
    def max_connection(self) -> int:
        """ The maximal number of connections that the pool can have. """
        return self._maxconn

    @max_connection.setter
    def max_connection(self, val) -> None:
        """ The maximal number of connections that the pool can have. """
        if val < self._minconn:
            raise ValueError("The maxconn must be greater than minconn.")
        self._maxconn = val


class ThreadedConnectionPool(ConnectionPool):
    """
    A connection pool that can be shared between threads. It's inherited from 
    :class:`bonsai.pool.ConnectionPool`.

    :param LDAPClient client: the :class:`bonsai.LDAPClient` that's used to create
                connections.
    :param int minconn: the minimum number of connections that's created
                after the pool is opened.
    :param int maxconn: the maximum number of connections in the pool.
    :param bool block: when it's True, the get method will block when no
                connection is available in the pool.
    :param \*\*kwargs: additional keyword arguments that are passed to
                the :meth:`bonsai.LDAPClient.connect` method.
    :raises ValueError: when the minconn is negative or the maxconn is less
        than the minconn.
    """
    def __init__(
        self,
        client: "LDAPClient",
        minconn: int = 1,
        maxconn: int = 10,
        block: bool = True,
        **kwargs: Dict[str, Any]
    ) -> None:
        """ Init method. """
        super().__init__(client, minconn, maxconn, **kwargs)
        self._block = block
        self._lock = threading.Condition()

    def get(self, timeout: Optional[float] = None):
        """
        Get a connection from the connection pool.

        :param float timeout: a timeout until waiting for free connection.
        :raises EmptyPool: when the pool is empty.
        :raises ClosedPool: when the method is called on a closed pool.
        :return: an LDAP connection object.
        """
        self._lock.acquire()
        try:
            if self._block:
                self._lock.wait_for(lambda: not self.empty or self._closed, timeout)
            conn = super().get()
            self._lock.notify()
            return conn
        finally:
            self._lock.release()

    def put(self, conn) -> None:
        self._lock.acquire()
        try:
            super().put(conn)
            self._lock.notify()
        finally:
            self._lock.release()

    def close(self) -> None:
        self._lock.acquire()
        try:
            super().close()
            self._lock.notify_all()
        finally:
            self._lock.release()



import asyncio
import sys

from ..ldapconnection import BaseLDAPConnection, LDAPSearchScope
from ..errors import LDAPError, NotAllowedOnNonleaf

# Backwards compatibility from 3.5.
if sys.version_info.minor < 5:
    StopAsyncIteration = StopIteration


class AIOLDAPConnection(BaseLDAPConnection):
    """
    Asynchronous LDAP connection object that works with asyncio.
    It has the same methods and properties as :class:`bonsai.LDAPConnection`, but
    with the exception of :meth:`bonsai.LDAPConnection.close` and
    :meth:`bonsai.LDAPConnection.fileno` all of them are awaitable.

    :param LDAPClient client: a client object.
    :param loop: an asyncio IO loop.
    """

    def __init__(self, client, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self.__open_coro = None
        super().__init__(client, is_async=True)

    async def __aenter__(self):
        """ Async context manager entry point. """
        return await self.__open_coro

    async def __aexit__(self, type, value, traceback):
        """ Async context manager exit point. """
        self.close()

    def __await__(self):
        return self.__open_coro.__await__()  # Hack to avoid returning a coroutine.

    __iter__ = __await__

    def _ready(self, msg_id, fut):
        self._loop.remove_reader(self.fileno())
        self._loop.remove_writer(self.fileno())
        try:
            res = super().get_result(msg_id)
            if res is not None:
                fut.set_result(res)
            else:
                self._loop.add_reader(self.fileno(), self._ready, msg_id, fut)
                self._loop.add_writer(self.fileno(), self._ready, msg_id, fut)
        except LDAPError as exc:
            fut.set_exception(exc)

    async def _poll(self, msg_id, timeout=None):
        fut = asyncio.Future()
        self._loop.add_reader(self.fileno(), self._ready, msg_id, fut)
        self._loop.add_writer(self.fileno(), self._ready, msg_id, fut)
        try:
            return await asyncio.wait_for(fut, timeout)
        except Exception as exc:
            self._loop.remove_reader(self.fileno())
            self._loop.remove_writer(self.fileno())
            raise exc

    def _evaluate(self, msg_id, timeout=None):
        return self._poll(msg_id, timeout)

    def open(self, timeout=None):
        self.__open_coro = super().open(timeout)
        return self

    async def delete(self, dname, timeout=None, recursive=False):
        try:
            return await super().delete(dname, timeout, recursive)
        except NotAllowedOnNonleaf as exc:
            if recursive:
                results = await self.search(
                    dname, LDAPSearchScope.ONELEVEL, attrlist=["1.1"], timeout=timeout
                )
                for res in results:
                    await self.delete(res.dn, timeout, True)
                return await self.delete(dname, timeout, False)
            else:
                raise exc

    async def _search_iter_anext(self, search_iter):
        try:
            return next(search_iter)
        except StopIteration:
            msgid = search_iter.acquire_next_page()
            if msgid is None:
                raise StopAsyncIteration
            search_iter = await self._evaluate(msgid)
            return next(search_iter)



import pytest

from bonsai import LDAPURL
from bonsai import LDAPDN
from bonsai.errors import InvalidDN


@pytest.fixture
def valid_ldapurl():
    """ Get a valid LDAPURL. """
    strurl = "ldaps://testurl:444/cn=test,dc=test?sn,gn?base?(objectclass=*)?1.2.3.4"
    return LDAPURL(strurl)


def test_get_address(valid_ldapurl):
    """ Test get_address method. """
    ldapi_url = LDAPURL("ldapi://%2Ftmp%2Fldapi")
    assert valid_ldapurl.get_address() == "ldaps://testurl:444"
    assert ldapi_url.get_address() == "ldapi://%2Ftmp%2Fldapi"


def test_get_host_properties(valid_ldapurl):
    """ Test getting LDAPURL host properties. """
    ldapi_url = LDAPURL("ldapi://%2Ftmp%2Fldapi")
    assert valid_ldapurl.scheme == "ldaps"
    assert valid_ldapurl.host == "testurl"
    assert valid_ldapurl.port == 444
    assert ldapi_url.scheme == "ldapi"
    assert ldapi_url.port == 0


def test_set_host_properties():
    """ Test setting LDAPURL host properties. """
    url = LDAPURL()
    with pytest.raises(ValueError):
        url.host = ":malformed,@äđĐ-"
    with pytest.raises(ValueError):
        url.port = "9922"
    with pytest.raises(ValueError):
        url.scheme = "http"

    url.host = "testurl2"
    url.port = 589
    url.scheme = "ldap"
    assert url.scheme == "ldap"
    assert url.host == "testurl2"
    assert url.port == 589


def test_get_bind_properties(valid_ldapurl):
    """ Test getting LDAPURL bind properties. """
    assert valid_ldapurl.basedn == LDAPDN("cn=test,dc=test")
    assert valid_ldapurl.scope == "base"
    assert valid_ldapurl.filter_exp == "(objectclass=*)"
    assert valid_ldapurl.attributes == ["sn", "gn"]


def test_set_bind_properties():
    """ Test setting LDAPURL bind properties. """
    url = LDAPURL()
    with pytest.raises(InvalidDN):
        url.basedn = "test"

    url.basedn = LDAPDN("cn=test")
    assert str(url.basedn) == "cn=test"


def test_str(valid_ldapurl):
    """ Test __str__ method of LDAPURL. """
    assert (
        str(valid_ldapurl)
        == "ldaps://testurl:444/cn=test,dc=test?sn,gn?base?(objectclass=*)?1.2.3.4"
    )
    assert str(LDAPURL("ldap://127.0.0.1/cn=x?cn")) == "ldap://127.0.0.1:389/cn=x?cn"
    assert str(LDAPURL("ldap:///")) == "ldap://localhost:389"
    assert str(LDAPURL("ldapi:///")) == "ldapi://localhost"
    assert not LDAPURL("ldap:///") == "http://localhost:389"
    assert "<LDAPURL" in repr(valid_ldapurl)


def test_conversion():
    """ Test ValueError exception for invalid URL format. """
    with pytest.raises(ValueError):
        _ = LDAPURL("ldap://failed.com/?falsedn?d")


def test_del_attr(valid_ldapurl):
    """ Test trying to delete an attribute. """
    with pytest.raises(AttributeError):
        del valid_ldapurl.host
    try:
        _ = valid_ldapurl.host
    except AttributeError:
        pytest.fail("Attribute not should be deleted.")


def test_invalid():
    """ Test invalid LDAP URLs. """
    with pytest.raises(ValueError):
        _ = LDAPURL("http://localhost")
    with pytest.raises(ValueError):
        _ = LDAPURL("ldaps://localost.")


def test_scope():
    """ Test scope and scope_num property. """
    url = LDAPURL("ldap:///??one")
    assert url.scope_num == 1
    url.scope = "base"
    assert url.scope_num == 0

    with pytest.raises(TypeError):
        url.scope = 2.1
    with pytest.raises(ValueError):
        url.scope = "all"


def test_ipv6():
    """ Test IPv6 address """
    url = LDAPURL(
        "ldap://[2001:db8:85a3::8a2e:370:7334]:1498/"
        "o=University%20of%20Michigan,c=US??one?"
        "(cn=Babs%20Jensen)"
    )
    assert url.host == "2001:db8:85a3::8a2e:370:7334"
    assert url.port == 1498
    assert url.scope == "one"
    assert url.filter_exp == "(cn=Babs Jensen)"
    addr = url.get_address()
    assert addr == "ldap://[2001:db8:85a3::8a2e:370:7334]:1498"
    with pytest.raises(ValueError):
        _ = LDAPURL("ldap://2001::85::37:7334")



import pytest
import base64
import os

from io import StringIO, BytesIO
from bonsai import LDIFReader, LDIFError


def test_init_params():
    """ Test constructor parameters for LDIFReader. """
    with pytest.raises(TypeError):
        _ = LDIFReader("wrong")
    with pytest.raises(TypeError):
        _ = LDIFReader(StringIO(), max_length=None)
    with pytest.raises(TypeError):
        _ = LDIFReader(BytesIO())
    inp = StringIO()
    ldif = LDIFReader(inp, max_length=100)
    assert ldif.input_file == inp
    assert ldif.max_length == 100


def test_version():
    """ Test setting version attribute from LDIF. """
    text = "version: 1\ndn: cn=test\ncn: test\n"
    with StringIO(text) as ldif:
        reader = LDIFReader(ldif)
        ent = next(reader)
        assert reader.version == 1
        assert ent.dn == "cn=test"


def test_missing_dn():
    """ Test missing distinguished name in LDIF entry. """
    text = "changetype: add\nsn: test\ncn: test\n"
    with StringIO(text) as ldif:
        reader = LDIFReader(ldif)
        with pytest.raises(LDIFError) as err:
            _ = next(reader)
        assert "Missing distinguished name" in str(err)
        assert "entry #1" in str(err)


def test_invalid_file():
    """ Test invalid and too long line. """
    text = " invalid\n"
    with StringIO(text) as ldif:
        reader = LDIFReader(ldif)
        with pytest.raises(LDIFError) as err:
            _ = next(reader)
        assert "Parser error" in str(err)
    text = "dn: cn=test\nnotvalid attribute\n"
    with StringIO(text) as ldif:
        reader = LDIFReader(ldif)
        with pytest.raises(LDIFError) as err:
            _ = next(reader)
        assert "Invalid attribute value pair:" in str(err)
        assert "entry #1" in str(err)
    text = "dn: cn=toolong\n"
    with StringIO(text) as ldif:
        reader = LDIFReader(ldif, max_length=12)
        with pytest.raises(LDIFError) as err:
            _ = next(reader)
        assert "too long" in str(err)
        assert "Line 1" in str(err)
    text = "dn: cn=test notvalid: attribute\n"
    with StringIO(text) as ldif:
        reader = LDIFReader(ldif)
        with pytest.raises(LDIFError) as err:
            _ = next(reader)
        assert "Invalid attribute value pair:" in str(err)
        assert "entry #1" in str(err)


def test_comment():
    """ Test parsing comment lines in LDIF files. """
    ldif = "# DN: cn=test\ndn: cn=test\n#Other comment line.\ncn: test\n"
    with StringIO(ldif) as test:
        reader = LDIFReader(test)
        ent = next(reader)
        assert ent.dn == "cn=test"
        assert ent["cn"] == ["test"]
    multiline = "# A long multiline comment\n in an LDIF file.\ndn: cn=test\n"
    with StringIO(multiline) as test:
        reader = LDIFReader(test)
        ent = next(reader)
        assert ent.dn == "cn=test"


def test_input_file():
    """ Test input_file property. """
    inp = StringIO()
    ldif = LDIFReader(inp)
    assert ldif.input_file == inp
    with pytest.raises(TypeError):
        ldif.input_file = None
    inp2 = StringIO()
    ldif.input_file = inp2
    assert ldif.input_file == inp2


def test_autoload():
    """ Test autoload property. """
    inp = StringIO()
    ldif = LDIFReader(inp)
    assert ldif.autoload == True
    with pytest.raises(TypeError):
        ldif.autoload = "Yes"
    ldif.autoload = False
    assert ldif.autoload == False


def test_resource_handlers():
    """ Test resource_handlers property. """
    inp = StringIO()
    ldif = LDIFReader(inp)
    assert isinstance(ldif.resource_handlers, dict)
    assert "file" in ldif.resource_handlers.keys()
    with pytest.raises(ValueError):
        ldif.resource_handlers = {"New": "dict"}
    ldif.resource_handlers["http"] = lambda x: x
    assert "http" in ldif.resource_handlers.keys()


def test_multiline_attribute():
    """ Test parsing multiline attributes in LDIF. """
    text = "dn: cn=unimaginably+sn=very,ou=very,dc=very,dc=long,\n dc=line\ncn: unimaginably\nsn: very\nsn: long\n"
    with StringIO(text) as test:
        reader = LDIFReader(test)
        ent = next(reader)
    assert ent.dn == "cn=unimaginably+sn=very,ou=very,dc=very,dc=long,dc=line"
    assert ent["cn"][0] == "unimaginably"
    assert ent["sn"][0] == "very"
    assert ent["sn"][1] == "long"


def test_multiple_entries():
    """ Test parsing multiple entries in one LDIF. """
    text = "dn: cn=test1\ncn: test1\n\ndn: cn=test2\ncn: test2\n"
    with StringIO(text) as test:
        reader = LDIFReader(test)
        entries = list(reader)
    assert len(entries) == 2
    assert entries[0].dn == "cn=test1"
    assert entries[1]["cn"][0] == "test2"


def test_encoded_attributes():
    """ Test parsing base64 encoded attributes. """
    attr = "test"
    text = "version: 1\ndn: cn=test\ncn:: {0}\n".format(
        base64.b64encode(attr.encode("UTF-8")).decode("UTF-8")
    )
    with StringIO(text) as test:
        reader = LDIFReader(test)
        ent = next(reader)
    assert ent.dn == "cn=test"
    assert ent["cn"][0] == attr


def test_load_resource():
    """ Test load_resource method. """
    curdir = os.path.abspath(os.path.dirname(__file__))
    with StringIO() as test:
        test.name = "dummy"
        reader = LDIFReader(test)
        with pytest.raises(LDIFError) as err:
            reader.load_resource("ftp://dummy.com")
            assert "Unsupported URL format" in str(err)
        with pytest.raises(LDIFError) as err:
            reader.load_resource("ftp//dummy.com")
            assert "Unsupported URL format" in str(err)
        url = "file://{0}".format(
            os.path.join(curdir, os.path.join("testenv", "test.jpeg"))
        )
        content = reader.load_resource(url)
        assert len(content) != 0
        assert isinstance(content, bytes)


def test_url_attribute():
    """ Test URL attribute in LDIF file. """
    text = "dn: cn=test\ncn: test1\njpegPhoto:< file://./testenv/test.jpeg\n"
    with StringIO(text) as test:
        test.name = __file__
        reader = LDIFReader(test)
        ent = next(reader)
    assert ent.dn == "cn=test"
    assert len(ent["jpegPhoto"][0]) == 1959
    assert isinstance(ent["jpegPhoto"][0], bytes)


def test_changetype():
    """ Test changetype attribute in LDIF file. """
    text = "dn: cn=test\nchangetype: add\ncn: test\n"
    with StringIO(text) as test:
        reader = LDIFReader(test)
        ent = next(reader)
    assert ent.dn == "cn=test"
    assert "cn" in ent
    assert "changetype" not in ent


def test_missing_attribute():
    """ Test missing attribute in LDIF-CHANGE. """
    text = "dn: cn=test\nchangetype: modify\nadd: sn\ncn: test\n"
    with StringIO(text) as test:
        reader = LDIFReader(test)
        with pytest.raises(LDIFError):
            _ = next(reader)


def test_modify_change():
    """ Test loading modified attributes from LDIF-CHANGE. """
    text = """dn: cn=test
changetype: modify
add: sn
sn: testing1
sn: testing2
-
replace: uid
uid: tester
-
delete: gidNumber
-
delete: objectclass
objectClass: posixUser

"""
    with StringIO(text) as test:
        reader = LDIFReader(test)
        ent = next(reader)
        status = ent._status()
        assert status["sn"]["@status"] == 1
        assert status["sn"]["@added"] == ["testing1", "testing2"]
        assert status["sn"]["@deleted"] == []
        assert status["uid"]["@status"] == 2
        assert status["uid"]["@added"] == ["tester"]
        assert status["uid"]["@deleted"] == []
        assert status["objectClass"]["@status"] == 1
        assert status["objectClass"]["@added"] == []
        assert status["objectClass"]["@deleted"] == ["posixUser"]
        assert status["@deleted_keys"] == ["gidNumber"]



from .ldapdn import LDAPDN
from .ldapurl import LDAPURL
from .ldapconnection import LDAPConnection
from .ldapconnection import LDAPSearchScope
from .ldapentry import LDAPEntry
from .ldapentry import LDAPModOp
from .ldapclient import LDAPClient
from .ldapreference import LDAPReference
from .ldapvaluelist import LDAPValueList
from .ldif import LDIFError, LDIFReader, LDIFWriter
from .errors import *
from .utils import *

__version__ = "1.1.0"



from typing import Any, List, Union, Iterable, Optional

import bonsai


class LDAPValueList(list):
    """
    Modified list that tracks added and deleted values. It also contains
    only unique elements. The elements are compared to their lower-cased
    string representations.

    A new LDAPValueList can be created optionally from an existing
    sequence object.

    :param items: a sequence object.
    :raises ValueError: if `items` has a non-unique element.
    """

    __slots__ = ("__deleted", "__added", "__status")

    def __init__(self, items: Optional[Iterable[Any]] = None) -> None:
        super().__init__()
        self.__added = []  # type: List[str]
        self.__deleted = []  # type: List[str]
        self.__status = 0
        if items:
            for item in items:
                self.append(item)

    @staticmethod
    def __balance(lst1: List[str], lst2: List[str], value: Any) -> None:
        """
        Balancing between the two list (__added, __deleted),
        make sure the same element is not in both lists.
        """
        try:
            lst1.remove(value)
        except ValueError:
            lst2.append(value)

    def _append_unchecked(self, value: Any) -> None:
        super().append(value)

    def _remove_unchecked(self, value: Any) -> None:
        try:
            super().remove(value)
        except ValueError:
            pass

    @property
    def _status_dict(self) -> dict:
        return {
            "@status": self.__status,
            "@added": self.__added.copy(),
            "@deleted": self.__deleted.copy(),
        }

    @_status_dict.setter
    def _status_dict(self, value: Any) -> None:
        raise TypeError("Can not change _status_dict")

    def __contains__(self, item: Any) -> bool:
        return bonsai.utils._unique_contains(self, item)[0]

    def __delitem__(self, idx: Union[int, slice]) -> None:
        old_value = super().__getitem__(idx)
        if type(idx) == slice:
            for item in old_value:
                self.__balance(self.__added, self.__deleted, item)
        else:
            self.__balance(self.__added, self.__deleted, old_value)
        super().__delitem__(idx)

    def __mul__(self, value: Any) -> "LDAPValueList":
        raise TypeError("Cannot multiple LDAPValueList.")

    def __add__(self, other: Iterable) -> "LDAPValueList":
        if type(other) != list and type(other) != LDAPValueList:
            raise TypeError("Can only concatenate list and LDAPValueList.")
        new_list = self.copy()
        new_list.extend(other)
        return new_list

    def __iadd__(self, other: Iterable) -> "LDAPValueList":
        if type(other) != list and type(other) != LDAPValueList:
            raise TypeError("Can only concatenate list and LDAPValueList.")
        self.extend(other)
        return self

    def __setitem__(self, idx: Union[int, slice], value: Any) -> None:
        old_value = self[idx]
        if type(idx) == slice:
            for item in value:
                if item in self:
                    raise ValueError("%r is already in the list." % item)
            for item in old_value:
                self.__balance(self.__added, self.__deleted, item)
            for item in value:
                self.__balance(self.__deleted, self.__added, item)
        else:
            if value in self:
                raise ValueError("%r is already in the list." % value)
            self.__balance(self.__added, self.__deleted, old_value)
            self.__balance(self.__deleted, self.__added, value)
        super().__setitem__(idx, value)

    def append(self, item: Any) -> None:
        """
        Add a unique item to the end of the LDAPValueList.

        :param item: New item.
        :raises ValueError: if the `item` is not unique.
        """

        if item in self:
            raise ValueError("%r is already in the list." % item)
        self.__balance(self.__deleted, self.__added, item)
        self.__status = 1
        super().append(item)

    def extend(self, items: Iterable[Any]) -> None:
        """
        Extend the LDAPValueList by appending all the items in the given
        list. All element in `items` must be unique and also not
        represented in the LDAPValueList.

        :param items: List of new items.
        :raises ValueError: if any of the items is already in the list.
        """
        for item in items:
            if item in self:
                raise ValueError("%r is already in the list." % item)
        for item in items:
            self.__balance(self.__deleted, self.__added, item)
        self.__status = 1
        super().extend(items)

    def insert(self, idx: int, value: Any) -> None:
        """
        Insert a unique item at a given position.

        :param int idx: the position.
        :param value: the new item.
        :raises ValueError: if the `item` is not unique.
        """
        if value in self:
            raise ValueError("%r is already in the list." % value)
        self.__balance(self.__deleted, self.__added, value)
        self.__status = 1
        super().insert(idx, value)

    def remove(self, value: Any) -> None:
        """
        Remove the first item from the LDAPValueList whose value is `value`.

        :param value: the item to be removed.
        :raises ValueError: if `value` is not int the list.
        """
        contain, obj = bonsai.utils._unique_contains(self, value)
        if not contain:
            raise ValueError("%r is not in the list." % value)
        super().remove(obj)
        self.__status = 1
        self.__balance(self.__added, self.__deleted, obj)

    def pop(self, idx: int = -1) -> Any:
        """
        Remove the item at the given position in the LDAPValueList, and
        return it. If no index is specified, pop() removes and returns the
        last item in the list.

        :param int idx: optional index.
        """
        value = super().pop(idx)
        self.__balance(self.__added, self.__deleted, value)
        self.__status = 1
        return value

    def clear(self) -> None:
        """ Remove all items from the LDAPValueList. """
        del self[:]

    def copy(self) -> "LDAPValueList":
        """
        Return a shallow copy of the LDAPValueList. This includes
        the status and the previously added and deleted items.

        :rtype: LDAPValueList
        :return: The copy of the LDAPValueList.
        """
        new_list = LDAPValueList()
        for item in self:
            new_list._append_unchecked(item)
        new_list.__added = self.__added.copy()
        new_list.__deleted = self.__deleted.copy()
        new_list.__status = self.__status
        return new_list

    @property
    def added(self) -> List:
        """ List of the added values. """
        return self.__added

    @added.setter
    def added(self, value: Any) -> None:
        raise ValueError("Added attribute cannot be changed.")

    @property
    def deleted(self) -> List:
        """ List of the deleted values. """
        return self.__deleted

    @deleted.setter
    def deleted(self, value: Any) -> None:
        raise ValueError("Deleted attribute cannot be changed.")

    @property
    def status(self) -> int:
        """
        The status of the LDAPValueList. The status can be:
            - 0: unchanged.
            - 1: added or deleted item to list.
            - 2: replaced the entire list.
        """
        return self.__status

    @status.setter
    def status(self, value: int) -> None:
        if type(value) != int:
            raise TypeError("Status must be int.")
        if value > -1 and value < 3:
            self.__status = value
        else:
            raise ValueError("Status must be between 0 and 2.")



from typing import Any, Optional, Union
from gevent.socket import wait_readwrite

from ..ldapconnection import BaseLDAPConnection, LDAPSearchScope
from ..ldapdn import LDAPDN
from ..errors import NotAllowedOnNonleaf

MYPY = False

if MYPY:
    from ..ldapclient import LDAPClient

class GeventLDAPConnection(BaseLDAPConnection):
    """
    Asynchronous LDAP connection object that works with Gevent.
    It has the same methods and properties as :class:`bonsai.LDAPConnection`.

    :param LDAPClient client: a client object.
    """

    def __init__(self, client: 'LDAPClient') -> None:
        super().__init__(client, is_async=True)

    def _poll(self, msg_id: int, timeout: Optional[float] = None) -> Any:
        while True:
            res = self.get_result(msg_id)
            if res is not None:
                return res
            wait_readwrite(self.fileno(), timeout=timeout)

    def _evaluate(self, msg_id: int, timeout: Optional[float] = None) -> Any:
        return self._poll(msg_id, timeout)

    def delete(self, dname: Union[str, LDAPDN], timeout: Optional[float] = None,
               recursive: bool = False) -> bool:
        try:
            return super().delete(dname, timeout, recursive)
        except NotAllowedOnNonleaf as exc:
            if recursive:
                results = self.search(dname, LDAPSearchScope.ONELEVEL,
                                      attrlist=['1.1'], timeout=timeout)
                for res in results:
                    self.delete(res.dn, timeout, True)
                return self.delete(dname, timeout, False)
            else:
                raise exc



import pytest

from bonsai import LDAPDN
from bonsai import errors

VALID_STRDN = "cn=user,dc=test,dc=local"


@pytest.fixture
def dnobj():
    """ Get a LDAPDN object with valid DN. """
    return LDAPDN(VALID_STRDN)


def test_rdn(dnobj):
    """ Test methods for retrieving and changing RDNs. """
    assert dnobj.rdns[0] == (("cn", "user"),)
    assert dnobj[0] == "cn=user"
    assert dnobj[1:] == "dc=test,dc=local"
    dnobj[1:] = "dc=test2"
    assert dnobj == "cn=user,dc=test2"
    with pytest.raises(IndexError):
        _ = dnobj[7]
    with pytest.raises(TypeError):
        _ = dnobj["test"]
    with pytest.raises(ValueError):
        dnobj.rdns = (("dc", "test"),)


def test_str(dnobj):
    """ Test __str__ method of LDAPDN object. """
    assert str(dnobj) == VALID_STRDN


def test_emptydn():
    """ Test empty distinguished name. """
    empty = LDAPDN("")
    assert empty[1:] == ""


def test_equal(dnobj):
    """ Test __eq__ method of LDAPDN object. """
    assert dnobj == LDAPDN(VALID_STRDN)
    assert dnobj == LDAPDN(VALID_STRDN.title())
    assert dnobj == VALID_STRDN.upper()


def test_invaliddn():
    """ Test InvalidDN exception. """
    with pytest.raises(errors.InvalidDN):
        _ = LDAPDN("cn=test,dc=one+two")


def test_special_char():
    """ Test parsing special characters in DN string. """
    spec = LDAPDN(r"cn=special\, name,dc=test,dc=local")
    assert str(spec) == r"cn=special\, name,dc=test,dc=local"


def test_setitem():
    """ Test setting RDNs for DN object. """
    dnobj = LDAPDN("sn=some+gn=thing,dc=test,dc=local")
    assert "sn=some+gn=thing" == dnobj[0]
    dnobj[0] = "cn=user"
    assert "cn=user,dc=test,dc=local" == dnobj
    dnobj[1] = "ou=group1,ou=group2"
    assert "cn=user,ou=group1,ou=group2,dc=local" == dnobj
    dnobj[2:] = "dc=local"
    assert "cn=user,ou=group1,dc=local" == dnobj

    with pytest.raises(TypeError):
        dnobj["invalid"] = "ou=group1,ou=group2"
    with pytest.raises(ValueError):
        dnobj[0] = 3
    with pytest.raises(errors.InvalidDN):
        dnobj[1] = "test,group"


def test_repr(dnobj):
    """ Test representation. """
    assert "<LDAPDN" in repr(dnobj)



from ipaddress import IPv6Address
from typing import Any, List, Union, Tuple, Optional

import re
import urllib.parse
from .ldapdn import LDAPDN


class LDAPURL:
    """
    LDAP URL object for handling LDAP connection informations, such as
    hostname, port, LDAP bind DN, search attributes, scope (base,sub or
    one) filter, and extensions. If `strurl` is None, then the default
    url is `ldap://localhost:389`.

    :param str strurl: string representation of a valid LDAP URL. Must \
    be started with `ldap://`, `ldaps://` or `ldapi://`.

    :raises ValueError: if the string parameter is not a valid LDAP URL.
    """

    __slots__ = ("__hostinfo", "__searchinfo", "__extensions", "__ipv6")

    def __init__(self, strurl: Optional[str] = None) -> None:
        """ Init method. """
        self.__hostinfo = ("ldap", "localhost", 389)  # type: Tuple[str, str, int]
        # Default values to the search parameters.
        self.__searchinfo = (
            LDAPDN(""),
            [],
            "",
            "",
        )  # type: Tuple[LDAPDN, List[str], str, str]
        self.__extensions = []  # type: List[str]
        self.__ipv6 = False
        if strurl:
            self.__str2url(strurl)

    def __delattr__(self, attr: str) -> None:
        """ None of the attributes can be deleted. """
        raise AttributeError("%s cannot be deleted." % attr)

    def __str2url(self, strurl: str) -> None:
        """ Parsing string url to LDAPURL."""
        # RegExp for [ldap|ldaps|ldapi]://[host]:[port]/[basedn]?[attrs]?[scope]
        # ?[filter]?[exts]
        valid = re.compile(
            r"^(ldap[s|i]?)://(([^:/?]*)?([:]([1-9][0-9]{0,4}))?"
            r"|[\[]?([^/?\]]*)([\]][:]([1-9][0-9]{0,4}))?)[/]?"
            r"([^\]/:?]*)?[\?]?([^\]:?]*)?[\?]?([^\]:?]*)?[\?]?"
            r"([^\]:?]*)?[\?]?([^\]:?]*)?$",
            re.IGNORECASE,
        )
        scheme, host, port = self.__hostinfo
        binddn, attrlist, scope, filter_exp = self.__searchinfo
        match = valid.match(strurl)
        if match:
            scheme = match.group(1).lower()
            if scheme == "ldaps":
                port = 636
            elif scheme == "ldapi":
                port = 0
            # The hostname
            if match.group(3) or match.group(6):
                hostname = match.group(3) or match.group(6)
                if scheme != "ldapi":
                    if self.is_valid_hostname(hostname):
                        host = hostname
                        if match.group(6):
                            self.__ipv6 = True
                    else:
                        raise ValueError(
                            "'{0}' has an invalid hostname.".format(strurl)
                        )
                else:
                    host = hostname
            # The portnumber for IPv4
            if match.group(5):
                port = int(match.group(5))
            # The portnumber for IPv6
            if match.group(8):
                port = int(match.group(8))
            # The LDAP bind DN
            if match.group(9):
                binddn = LDAPDN(urllib.parse.unquote(match.group(9)))
            # Attributes
            if match.group(10):
                attrlist = match.group(10).split(",")
            # Scope (base/one/sub)
            if match.group(11):
                _scope = match.group(11).lower()
                if _scope not in ("base", "one", "sub"):
                    raise ValueError("Invalid scope type.")
                scope = _scope
            # Filter
            if match.group(12):
                filter_exp = urllib.parse.unquote(match.group(12))
            # Extensions
            if match.group(13):
                self.__extensions = match.group(13).split(",")
            self.__hostinfo = (scheme, host, port)
            self.__searchinfo = (binddn, attrlist, scope, filter_exp)
        else:
            raise ValueError("'%s' is not a valid LDAP URL." % strurl)

    @staticmethod
    def is_valid_hostname(hostname: str) -> bool:
        """ Validate a hostname. """
        hostname_regex = re.compile(
            r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]"
            r"*[a-zA-Z0-9])\.)*([A-Za-z0-9]|"
            r"[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$"
        )
        try:
            # Try parsing IPv6 address.
            IPv6Address(hostname)
            return True
        except ValueError:
            # Try IPv4 and standard hostname.
            if hostname_regex.match(hostname):
                return True
            return False

    @property
    def host(self) -> str:
        """ The hostname. """
        return self.__hostinfo[1]

    @host.setter
    def host(self, value: str) -> None:
        """ Setter for hostname. """
        # RegExp for valid hostname.
        if not self.is_valid_hostname(value):
            raise ValueError("'%s' is not a valid host name." % value)
        else:
            self.__hostinfo = (self.__hostinfo[0], value, self.__hostinfo[2])

    @property
    def port(self) -> int:
        """ The portnumber. """
        return self.__hostinfo[2]

    @port.setter
    def port(self, value: int) -> None:
        """ Setter for portnumber. """
        if isinstance(value, int) and (value > 0 and value < 65535):
            self.__hostinfo = (self.__hostinfo[0], self.__hostinfo[1], value)
        else:
            raise ValueError("Port must be an int between 1 and 65535.")

    @property
    def scheme(self) -> str:
        """ The URL scheme."""
        return self.__hostinfo[0]

    @scheme.setter
    def scheme(self, value: str) -> None:
        """ Setter for URL scheme."""
        # It must be ldap, ldaps or ldapi
        if isinstance(value, str) and value.lower() in ("ldap", "ldaps", "ldapi"):
            self.__hostinfo = (value.lower(), self.__hostinfo[1], self.__hostinfo[2])
        else:
            raise ValueError("Scheme only be 'ldap', 'ldaps' or 'ldapi'.")

    @property
    def basedn(self) -> LDAPDN:
        """ The LDAP distinguished name for binding. """
        return self.__searchinfo[0]

    @basedn.setter
    def basedn(self, value: Union[LDAPDN, str]) -> None:
        """ Setter for LDAP distinguished name for binding. """
        self.__searchinfo = (
            LDAPDN(str(value)),
            self.__searchinfo[1],
            self.__searchinfo[2],
            self.__searchinfo[3],
        )

    @property
    def attributes(self) -> List[str]:
        """ The searching attributes. """
        return self.__searchinfo[1]

    @property
    def scope(self) -> str:
        """ The searching scope. """
        return self.__searchinfo[2]

    @scope.setter
    def scope(self, value: str) -> None:
        """ Setter for searching scope. """
        if isinstance(value, str):
            if value.lower() in ("base", "one", "sub"):
                self.__searchinfo = (
                    self.__searchinfo[0],
                    self.__searchinfo[1],
                    value,
                    self.__searchinfo[3],
                )
            else:
                raise ValueError(
                    """Scope must be one of these:
                            'base', 'one', 'sub'."""
                )
        else:
            raise TypeError("Scope must be a string.")

    @property
    def scope_num(self) -> int:
        """ Return the searching scope number. """
        if self.scope == "base":
            return 0
        if self.scope == "one":
            return 1
        if self.scope == "sub":
            return 2
        else:
            return -1

    @property
    def filter_exp(self) -> str:
        """ The searching filter expression. """
        return self.__searchinfo[3]

    def get_address(self) -> str:
        """
        Return the full address of the host.
        """
        if self.scheme == "ldapi":
            return "{0}://{1}".format(self.__hostinfo[0], self.__hostinfo[1])
        if self.__ipv6:
            return "{0}://[{1}]:{2:d}".format(*self.__hostinfo)
        else:
            return "{0}://{1}:{2:d}".format(*self.__hostinfo)

    def __eq__(self, other: Any) -> bool:
        """
        Check equality of two LDAPURL or an LDAPURL and a string.
        """
        if isinstance(other, LDAPURL):
            return (
                self.scheme == other.scheme
                and self.host == other.host
                and self.port == other.port
                and self.basedn == other.basedn
                and self.scope == other.scope
                and self.filter_exp == other.filter_exp
                and self.attributes == other.attributes
            )
        elif isinstance(other, str):
            try:
                other = LDAPURL(other)
            except ValueError:
                return False
            return self == other
        else:
            return False

    def __str__(self) -> str:
        """ Returns the full format of LDAP URL. """
        strurl = self.get_address()
        strattrs = ""
        strexts = ""
        if self.__searchinfo[1]:
            strattrs = ",".join(self.__searchinfo[1])
        if self.__extensions:
            strexts = ",".join(self.__extensions)
        strbind = "?".join(
            (
                str(self.__searchinfo[0]),
                strattrs,
                self.__searchinfo[2],
                self.__searchinfo[3],
                strexts,
            )
        )
        # Remove unnecessary question marks at the end of the string.
        while strbind[-1] == "?":
            strbind = strbind[:-1]
            if not strbind:
                break
        if strbind:
            strurl = "%s/%s" % (strurl, strbind)
        return strurl

    def __repr__(self) -> str:
        """ The LDAPURL representation. """
        return "<LDAPURL %s>" % str(self)
